from __future__ import annotations

import argparse
import errno
import inspect
import json as jsonlib
import os
import sys
import textwrap
import urllib.parse
from functools import wraps
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from dissect.target.helpers import docs, keychain
from dissect.target.helpers.docs import get_docstring
from dissect.target.loader import LOADERS_BY_SCHEME
from dissect.target.plugin import (
    FunctionDescriptor,
    OSPlugin,
    Plugin,
    find_functions,
    get_external_module_paths,
    load,
    load_modules_from_paths,
)
from dissect.target.plugins.general.plugins import (
    _get_default_functions,
    generate_functions_json,
    generate_functions_overview,
)
from dissect.target.target import Target
from dissect.target.tools.logging import configure_logging

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from datetime import datetime


USAGE_FORMAT_TMPL = "{prog} -f {name}{usage}"


def configure_generic_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-K", "--keychain-file", type=Path, help="keychain file in CSV format")
    parser.add_argument("-Kv", "--keychain-value", help="passphrase, recovery key or key file path value")
    parser.add_argument("-L", "--loader", action="store", help="select a specific loader (i.e. vmx, raw)")
    parser.add_argument("-v", "--verbose", action="count", default=0, help="increase output verbosity")
    parser.add_argument("--version", action="store_true", help="print version")
    parser.add_argument("-q", "--quiet", action="store_true", help="do not output logging information")
    parser.add_argument(
        "--plugin-path",
        action="store",
        nargs="+",
        type=Path,
        help="a file or directory containing plugins and extensions",
    )


def process_generic_arguments(args: argparse.Namespace) -> None:
    configure_logging(args.verbose, args.quiet, as_plain_text=True)

    if args.version:
        try:
            print("dissect.target version " + version("dissect.target"))
        except PackageNotFoundError:
            print("unable to determine version")
        sys.exit(0)

    targets = args.targets if hasattr(args, "targets") else [args.target] if hasattr(args, "target") else []
    if targets and args.loader:
        targets = args_to_uri(targets, args.loader)

    if hasattr(args, "targets"):
        args.targets = targets
    elif hasattr(args, "target"):
        args.target = targets[0]

    if args.keychain_file:
        keychain.register_keychain_file(args.keychain_file)

    if args.keychain_value:
        keychain.register_wildcard_value(args.keychain_value)

    paths = get_external_module_paths(args.plugin_path or [])
    load_modules_from_paths(paths)


def configure_plugin_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-f", "--function", help="one or more comma separated functions to execute")
    parser.add_argument("-xf", "--excluded-functions", help="functions to exclude from execution", default="")
    parser.add_argument(
        "-l",
        "--list",
        action="store",
        nargs="?",
        const="",
        default=None,
        help="list (matching) available plugins and loaders",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="do not execute the functions, but just print which functions would be executed",
    )


def process_plugin_arguments(parser: argparse.ArgumentParser, args: argparse.Namespace, rest: list[str]) -> set:
    """Processes the arguments concerting plugin functions (``-f``, ``--function``) and
    exclusion list (``-xf``, ``--exclude-funciton``).

    It puts the excluded function paths inside ``args.excluded_functions`` as a side effect.

    Returns:
        ``True`` if there are multiple output types detected, false otherwise.
    """

    # Show help for a function or in general
    if "-h" in rest or "--help" in rest:
        found_functions, _ = find_functions(args.function)
        if not len(found_functions):
            parser.error("function(s) not found, see -l for available plugins")

        func = found_functions[0]
        plugin_class = load(func)
        if issubclass(plugin_class, OSPlugin):
            obj = getattr(OSPlugin, func.method_name)
        else:
            obj = getattr(plugin_class, func.method_name)

        if isinstance(obj, type) and issubclass(obj, Plugin):
            parser = generate_argparse_for_plugin_class(obj, usage_tmpl=USAGE_FORMAT_TMPL)
        elif isinstance(obj, (Callable, property)):
            parser = generate_argparse_for_method(getattr(obj, "fget", obj), usage_tmpl=USAGE_FORMAT_TMPL)
        else:
            parser.error(f"can't find plugin with function `{func.method_name}`")
        parser.print_help()
        parser.exit(0)

    # Show the list of available plugins for the given optional target and optional
    # search pattern, only display plugins that can be applied to ANY targets
    if args.list is not None:
        list_plugins(args.targets, args.list, getattr(args, "children", False), getattr(args, "json", False))
        parser.exit(0)

    if not args.function:
        parser.error("argument -f/--function is required")

    funcs, invalid_funcs = find_functions(args.function)
    if any(invalid_funcs):
        parser.error(f"argument -f/--function contains invalid plugin(s): {', '.join(invalid_funcs)}")

    excluded_funcs, invalid_excluded_funcs = find_functions(args.excluded_functions)
    if any(invalid_excluded_funcs):
        parser.error(
            f"argument -xf/--excluded-functions contains invalid plugin(s): {', '.join(invalid_excluded_funcs)}",
        )

    args.excluded_functions = list({excluded.path for excluded in excluded_funcs})

    # Verify uniformity of output types, otherwise default to records.
    # Note that this is a heuristic, the targets are not opened yet because of
    # performance, so it might generate a false positive
    # (os.* on Windows includes other OS plugins),
    # however this is highly hypothetical, most plugins across OSes have
    # the same output types and most output types are records anyway.
    # Furthermore we really want the notification at the top, so this is the only
    # way forward. In the very unlikely case you have a
    # collection of non-record plugins that have record counterparts for
    # other OSes just refine the wildcard to exclude other OSes.
    # The only scenario that might cause this is with
    # custom plugins with idiosyncratic output across OS-versions/branches.
    return {func.output for func in funcs if func.path not in args.excluded_functions}


def open_targets(args: argparse.Namespace) -> Iterator[Target]:
    direct: bool = getattr(args, "direct", False)
    children: bool = getattr(args, "children", False)
    child: str | None = getattr(args, "child", None)
    targets: Iterable[Target] = (
        [Target.open_direct(args.targets)] if direct else Target.open_all(args.targets, children)
    )

    for target in targets:
        if child:
            try:
                target: Target = target.open_child(child)
            except Exception as e:
                target.log.exception("Exception while opening child %r: %s", child, e)  # noqa: TRY401
                target.log.debug("", exc_info=e)

        if getattr(args, "dry_run", False):
            print(f"Dry run on: {target}")

        yield target


def list_plugins(
    targets: list[str] | None = None,
    patterns: str = "",
    include_children: bool = False,
    json: bool = False,
) -> None:
    functions = None
    if targets:
        for target in Target.open_all(targets, include_children):
            functions, _ = find_functions(patterns or "*", target, compatibility=True, show_hidden=True)
    elif patterns:
        functions, _ = find_functions(patterns, Target(), show_hidden=True)
    else:
        functions = _get_default_functions()

    target = Target()
    fparser = generate_argparse_for_method(target.plugins, usage_tmpl=USAGE_FORMAT_TMPL)
    fargs, _ = fparser.parse_known_args()

    # Display in a user friendly manner
    if functions:
        if json:
            print('{"plugins": ', end="")
            print(generate_functions_json(functions), end="")
        else:
            print(generate_functions_overview(functions, include_docs=fargs.print_docs))
    elif json:
        print("{", end="")

    # No real targets specified, show the available loaders
    if not targets:
        fparser = generate_argparse_for_method(target.loaders, usage_tmpl=USAGE_FORMAT_TMPL)
        fargs, _ = fparser.parse_known_args()
        if json:
            print(', "loaders": ', end="")
        target.loaders(**vars(fargs))

    if json:
        print("}")


def generate_argparse_for_method(
    method: Callable,
    usage_tmpl: str | None = None,
) -> argparse.ArgumentParser:
    """Generate an ``argparse.ArgumentParser`` for a bound or unbound ``Plugin`` class method."""

    # allow functools.partial wrapped method
    while hasattr(method, "func"):
        method = method.func

    if not inspect.isfunction(method) and not inspect.ismethod(method):
        raise ValueError(f"Value `{method}` is not a plugin method")

    desc = method.__doc__ or docs.get_func_description(method, with_docstrings=True)

    if "\n" in desc:
        desc = inspect.cleandoc(desc)

    help_formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=desc, formatter_class=help_formatter, conflict_handler="resolve")

    _add_args_to_parser(parser, getattr(method, "__args__", []))

    usage = parser.format_usage()
    offset = usage.find(parser.prog) + len(parser.prog)
    func_name = method.__name__
    usage_tmpl = usage_tmpl or "{prog} {usage}"
    parser.usage = usage_tmpl.format(prog=parser.prog, name=func_name, usage=usage[offset:])

    return parser


def generate_argparse_for_plugin_class(
    plugin_cls: type[Plugin],
    usage_tmpl: str | None = None,
) -> argparse.ArgumentParser:
    """Generate an ``argparse.ArgumentParser`` for a ``Plugin`` class."""

    if not isinstance(plugin_cls, type) or not issubclass(plugin_cls, Plugin):
        raise TypeError(f"`plugin_cls` must be a valid plugin class, not `{plugin_cls}`")

    method_name = plugin_cls.__namespace__
    desc = docs.get_plugin_overview(plugin_cls, with_plugin_desc=True, with_func_docstrings=True)

    help_formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=desc, formatter_class=help_formatter, conflict_handler="resolve")

    _add_args_to_parser(parser, getattr(plugin_cls.__call__, "__args__", []))

    usage = parser.format_usage()
    offset = usage.find(parser.prog) + len(parser.prog)
    usage_tmpl = usage_tmpl or "{prog} {usage}"

    parser.usage = usage_tmpl.format(prog=parser.prog, name=method_name, usage=usage[offset:])

    return parser


def generate_argparse_for_plugin(
    plugin_instance: Plugin,
    usage_tmpl: str | None = None,
) -> argparse.ArgumentParser:
    """Generate an ``argparse.ArgumentParser`` for a ``Plugin`` instance."""

    if not isinstance(plugin_instance, Plugin):
        raise TypeError(f"`plugin_instance` must be a valid plugin instance, not `{plugin_instance}`")

    if not hasattr(plugin_instance, "__namespace__"):
        raise ValueError(f"Plugin `{plugin_instance}` is not a namespace plugin and is not callable")

    return generate_argparse_for_plugin_class(plugin_instance.__class__, usage_tmpl=usage_tmpl)


def _add_args_to_parser(
    parser: argparse.ArgumentParser,
    fargs: list[tuple[list[str], dict[str, Any]]],
) -> None:
    groups = {}
    default_group_options = {"required": False}
    for args, kwargs in fargs:
        if "group" in kwargs:
            group_name = kwargs.pop("group")
            options = kwargs.pop("group_options") if "group_options" in kwargs else default_group_options
            if group_name not in groups:
                group = parser.add_mutually_exclusive_group(**options)
                groups[group_name] = group
            else:
                group = groups[group_name]

            group.add_argument(*args, **kwargs)
        else:
            parser.add_argument(*args, **kwargs)


def execute_function_on_target(
    target: Target,
    func: FunctionDescriptor,
    args: list[str] | None = None,
) -> tuple[str, Any]:
    """Execute function on provided target with optional provided arguments.

    If no explicit arguments are provided, they will be parsed from ``sys.argv``.
    """

    func_cls, func_obj = target.get_function(func.name)
    plugin_method, parser = plugin_function_with_argparser(func_obj)

    if parser:
        known_args, _ = parser.parse_known_args(args)
        value = plugin_method(**vars(known_args))
    elif isinstance(func_obj, property):
        value = func_obj.__get__(func_cls)
    else:
        value = func_obj

    output_type = getattr(plugin_method, "__output__", "default") if plugin_method else "default"
    return (output_type, value)


def plugin_function_with_argparser(
    target_attr: Plugin | Callable,
) -> tuple[Callable | None, argparse.ArgumentParser | None]:
    """Resolves which plugin function to execute, and creates the argument parser for said plugin."""
    plugin_method = None
    parser = None

    # goes first because plugins are callable
    if isinstance(target_attr, Plugin):
        plugin_obj = target_attr

        if not plugin_obj.__namespace__:
            raise ValueError(f"Plugin {plugin_obj} is not callable")

        plugin_method = plugin_obj.__call__
        parser = generate_argparse_for_plugin(plugin_obj)
    elif callable(target_attr):
        plugin_method = target_attr
        parser = generate_argparse_for_method(target_attr)
    return plugin_method, parser


def persist_execution_report(output_dir: Path, report_data: dict, timestamp: datetime) -> Path:
    timestamp = timestamp.strftime("%Y-%m-%d-%H%M%S")
    report_filename = f"target-report-{timestamp}.json"
    report_full_path = output_dir / report_filename
    report_full_path.write_text(jsonlib.dumps(report_data, sort_keys=True, indent=4))
    return report_full_path


def catch_sigpipe(func: Callable) -> Callable:
    """Catches ``KeyboardInterrupt`` and ``BrokenPipeError`` (``OSError 22`` on Windows)."""

    @wraps(func)
    def wrapper(*args, **kwargs) -> int:
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            print("Aborted!", file=sys.stderr)
            return 1
        except OSError as e:
            # Only catch BrokenPipeError or OSError 22
            if e.errno in (errno.EPIPE, errno.EINVAL):
                devnull = os.open(os.devnull, os.O_WRONLY)
                os.dup2(devnull, sys.stdout.fileno())
                return 1
            # Raise other exceptions
            raise

    return wrapper


def args_to_uri(targets: list[str], loader_name: str, args: list[str] | None = None) -> list[str]:
    """Converts argument-style ``-L`` to URI-style.

    Turns:
        ``target-query /evtxs/* -L log --log-hint="evtx" -f evtx``
    into:
        ``target-query "log:///evtxs/*?hint=evtx" -f evtx``

    For loaders providing ``@arg()`` arguments.
    """
    loader = LOADERS_BY_SCHEME.get(loader_name, None)

    parser = argparse.ArgumentParser(
        argument_default=argparse.SUPPRESS, description="\n".join(textwrap.wrap(get_docstring(loader)))
    )
    for load_arg in getattr(loader, "__args__", []):
        parser.add_argument(*load_arg[0], **load_arg[1])
    args = vars(parser.parse_known_args(args)[0])
    return [f"{loader_name}://{target}" + (("?" + urllib.parse.urlencode(args)) if args else "") for target in targets]


def find_and_filter_plugins(
    functions: str, target: Target, excluded_func_paths: set[str] | None = None
) -> Iterator[FunctionDescriptor]:
    # Keep a set of plugins that were already executed on the target.
    executed_plugins = set()
    excluded_func_paths = excluded_func_paths or set()

    func_defs, _ = find_functions(functions, target)

    for func_def in func_defs:
        if func_def.path in excluded_func_paths:
            continue

        # Avoid executing same plugin for multiple OSes (like hostname)
        if func_def.name in executed_plugins:
            continue

        executed_plugins.add(func_def.name)

        yield func_def


def escape_str(value: str) -> str:
    """Escape non-ASCII, unicode characters and bytes to a printable form."""
    return repr(value)[1:-1]



from __future__ import annotations

from uuid import UUID

from dissect.cstruct import cstruct

from dissect.target.plugins.os.windows.dpapi.crypto import (
    CipherAlgorithm,
    HashAlgorithm,
    crypt_session_key_type1,
    crypt_session_key_type2,
)

blob_def = """
struct DPAPIBlob {
    DWORD   dwVersion;
    char    provider[16];
    DWORD   mkVersion;
    char    guid[16];
    DWORD   flags;
    DWORD   descriptionLength;
    char    description[descriptionLength];
    DWORD   CipherAlgId;
    DWORD   keyLen;
    DWORD   saltLength;
    char    salt[saltLength];
    DWORD   strongLength;
    char    strong[strongLength];
    DWORD   CryptAlgId;
    DWORD   hashLen;
    DWORD   hmacLength;
    char    hmac[hmacLength];
    DWORD   cipherTextLength;
    char    cipherText[cipherTextLength];
    DWORD   signLength;
    char    sign[signLength];
};
"""

c_blob = cstruct().load(blob_def)


class Blob:
    """Represents a DPAPI blob."""

    def __init__(self, data: bytes):
        self._blob = c_blob.DPAPIBlob(data)

        self.version = self._blob.dwVersion
        self.provider = str(UUID(bytes_le=self._blob.provider))
        self.mkversion = self._blob.mkVersion
        self.guid = str(UUID(bytes_le=self._blob.guid))
        self.flags = self._blob.flags
        self.description = self._blob.description.decode("utf-16-le")
        self.cipher_algorithm = CipherAlgorithm.from_id(self._blob.CipherAlgId)
        self.key_len = self._blob.keyLen
        self.salt = self._blob.salt
        self.strong = self._blob.strong
        self.hash_algorithm = HashAlgorithm.from_id(self._blob.CryptAlgId)
        self.hash_len = self._blob.hashLen
        self.hmac = self._blob.hmac
        self.cipher_text = self._blob.cipherText

        # All the blob data between the version, provider and sign fields
        # TODO: Replace with future offsetof function in cstruct
        self.blob = data[c_blob.DPAPIBlob.lookup["mkVersion"].offset : -(self._blob.signLength + len(c_blob.DWORD))]
        self.sign = self._blob.sign

        self.clear_text = None
        self.decrypted = False
        self.sign_computed = None

    def decrypt(
        self,
        master_key: bytes,
        entropy: bytes | None = None,
        strong_password: str | None = None,
        smart_card_secret: bytes | None = None,
    ) -> bool:
        """Try to decrypt the blob with the given master key.

        Args:
            master_key: Decrypted master key value.
            entropy: Optional entropy for decrypting the blob.
            strong_password: Optional password for decrypting the blob.
            smart_card_secret: MS Next Gen Crypto secret (e.g. from PIN code).

        Returns:
            True if decryption is succesful, False otherwise.
        """
        if self.decrypted:
            return True

        if not master_key:
            raise ValueError("No master key provided to decrypt blob with")

        for algo in [crypt_session_key_type1, crypt_session_key_type2]:
            session_key = algo(
                master_key,
                self.salt,
                self.hash_algorithm,
                entropy=entropy,
                smart_card_secret=smart_card_secret,
                strong_password=strong_password,
            )
            key = self.cipher_algorithm.derive_key(session_key, self.hash_algorithm)
            self.clear_text = self.cipher_algorithm.decrypt(self.cipher_text, key)

            padding = self.clear_text[-1]
            if padding <= self.cipher_algorithm.block_length:
                self.clear_text = self.clear_text[:-padding]

            # Check against provided HMAC
            self.sign_computed = algo(
                master_key,
                self.hmac,
                self.hash_algorithm,
                entropy=entropy,
                smart_card_secret=smart_card_secret,
                verify_blob=self.blob,
            )

            self.decrypted = self.sign_computed == self.sign
            if self.decrypted:
                return True

        self.decrypted = False
        return self.decrypted

    def __repr__(self) -> str:
        s = [
            "DPAPI BLOB",
            f"\tversion      = {self.version}",
            f"\tprovider     = {self.provider}",
            f"\tmkey         = {self.guid}",
            f"\tflags        = {self.flags:#x}",
            f"\tdescr        = {self.description}",
            f"\tcipher_algo   = {self.cipher_algorithm!r}",
            f"\thash_algo     = {self.hash_algorithm!r}",
            f"\tsalt         = {self.salt.hex()}",
            f"\thmac         = {self.hmac.hex()}",
            f"\tcipher       = {self.cipher_text.hex()}",
            f"\tsign         = {self.sign.hex()}",
        ]
        if self.sign_computed is not None:
            s.append(f"\tsign_computed = {self.sign_computed.hex()}")
        if self.clear_text is not None:
            s.append(f"\tcleartext    = {self.clear_text!r}")
        return "\n".join(s)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualKey
from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIPlugin
from dissect.target.plugins.os.windows.dpapi.keyprovider.defaultpassword.winlogon import (
    WinlogonDefaultPasswordKeyProviderPlugin,
)

if TYPE_CHECKING:
    from dissect.target.helpers.regutil import VirtualHive
    from dissect.target.target import Target


def test_dpapi_keyprovider_winlogon(target_win: Target, hive_hklm: VirtualHive) -> None:
    """test if we find a Winlogon DefaultPassword entry in the registry."""

    winlogon_path = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"
    winlogon_key = VirtualKey(hive_hklm, winlogon_path)
    winlogon_key.add_value("DefaultPassword", "password")
    hive_hklm.map_key(winlogon_path, winlogon_key)

    target_win.add_plugin(DPAPIPlugin, check_compatible=False)
    target_win.add_plugin(WinlogonDefaultPasswordKeyProviderPlugin)

    key = next(target_win.dpapi.keyprovider.defaultpassword.winlogon())

    assert key == ("dpapi.keyprovider.defaultpassword.winlogon", "password")



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.util import stream

from dissect.target import filesystem
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

BLOCK_SIZE = 64 * 0x100000  # 64 MiB

NTFS_NEEDLE = b"\xeb\x52\x90NTFS    \x00"
EXTFS_NEEDLE = b"\xff\xff\x53\xef"
# The extfs needle is 54 kiB from the start of the extfs volume
EXTFS_NEEDLE_OFFSET = 0x36 * 1024
FS_NEEDLES = [NTFS_NEEDLE, EXTFS_NEEDLE]


def scrape_pos(
    fp: BinaryIO,
    needles: list[bytes],
    block_size: int = BLOCK_SIZE,
) -> Iterator[tuple[bytes, int]]:
    while True:
        file_pos = fp.tell()
        block = fp.read(block_size)

        if not block:
            break

        for needle in needles:
            block_pos = -1
            while True:
                block_pos = block.find(needle, block_pos + 1)
                if block_pos == -1:
                    break

                offset = file_pos + block_pos
                yield needle, offset


class PhobosLoader(Loader):
    """Load Phobos Ransomware files.

    References:
        - https://www.hhs.gov/sites/default/files/overview-phobos-ransomware.pdf
    """

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".eight"

    def map(self, target: Target) -> None:
        fh = self.path.open("rb")

        fs_idx = 0
        for needle, offset in scrape_pos(fh, FS_NEEDLES):
            cur_seek = fh.tell()

            try:
                if needle == NTFS_NEEDLE:
                    volume = stream.RelativeStream(fh, offset)
                    fs = filesystem.open(volume)
                    size = fs.ntfs.boot_sector.NumberSectors * fs.ntfs.sector_size
                elif needle == EXTFS_NEEDLE:
                    volume = stream.RelativeStream(fh, offset - EXTFS_NEEDLE_OFFSET)
                    fs = filesystem.open(volume)
                    size = fs.extfs.block_count * fs.extfs.block_size

                target.filesystems.add(fs)
                target.fs.mount(f"fs{fs_idx}", fs)
                fs_idx += 1

                fh.seek(offset + size)
            except Exception:
                fh.seek(cur_seek)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

INVENTORY_PATHS = [
    # Windows
    "AppData/Roaming/VMware/inventory.vmls",
    # Linux
    ".vmware/inventory.vmls",
]


def find_vm_inventory(target: Target) -> Iterator[TargetPath]:
    """Search for inventory.vmls files in user home folders.

    Does not support older vmAutoStart.xml or vmInventory.xml formats."""

    for user_details in target.user_details.all_with_home():
        for inv_path in INVENTORY_PATHS:
            if (inv_file := user_details.home_path.joinpath(inv_path)).exists():
                yield inv_file


class VmwareWorkstationChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from VMware Workstation VM inventory."""

    __type__ = "vmware_workstation"

    def __init__(self, target: Target):
        super().__init__(target)
        self.inventories = list(find_vm_inventory(target))

    def check_compatible(self) -> None:
        if not self.inventories:
            raise UnsupportedPluginError("No VMWare inventories found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for inv in self.inventories:
            for line in inv.open("rt"):
                line = line.strip()
                if not line.startswith("vmlist"):
                    continue

                key, _, value = line.partition("=")
                if not key.strip().endswith(".config"):
                    continue

                value = value.strip().strip('"')
                if value.startswith("folder") or not value:
                    continue

                yield ChildTargetRecord(
                    type=self.__type__,
                    path=self.target.fs.path(value.strip('"')),
                    _target=self.target,
                )



"""Dissect plugin system.

See ``dissect/target/plugins/general/example.py`` for an example plugin.
"""

from __future__ import annotations

import fnmatch
import functools
import importlib
import importlib.util
import logging
import os
import sys
import traceback
from dataclasses import dataclass, field
from itertools import chain, zip_longest
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

try:
    from typing import TypeAlias  # novermin
except ImportError:
    # COMPAT: Remove this when we drop Python 3.9
    TypeAlias = Any

from flow.record import Record, RecordDescriptor

import dissect.target.plugins.os.default as default
from dissect.target.exceptions import PluginError, PluginNotFoundError, UnsupportedPluginError
from dissect.target.helpers import cache
from dissect.target.helpers.fsutil import has_glob_magic
from dissect.target.helpers.record import EmptyRecord
from dissect.target.helpers.utils import StrEnum

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.helpers.record import ChildTargetRecord
    from dissect.target.target import Target

log = logging.getLogger(__name__)

MODULE_PATH = "dissect.target.plugins"
"""The base module path to the in-tree plugins."""

OS_MODULE_PATH = "dissect.target.plugins.os"

OUTPUTS = (
    "default",
    "record",
    "yield",
    "none",
)
"""The different output types supported by ``@export``."""


class OperatingSystem(StrEnum):
    ANDROID = "android"
    BSD = "bsd"
    CITRIX = "citrix-netscaler"
    ESXI = "esxi"
    FORTIOS = "fortios"
    IOS = "ios"
    LINUX = "linux"
    MACOS = "macos"
    OSX = "osx"  # legacy
    PROXMOX = "proxmox"
    UNIX = "unix"
    VYOS = "vyos"
    WINDOWS = "windows"


@dataclass(frozen=True, eq=True)
class PluginDescriptor:
    # COMPAT: Replace with slots=True when we drop Python 3.9
    __slots__ = ("exports", "findable", "functions", "module", "namespace", "path", "qualname")

    module: str
    qualname: str
    namespace: str
    path: str
    findable: bool
    functions: list[str]
    exports: list[str]

    @property
    def cls(self) -> type[Plugin]:
        return load(self)


@dataclass(frozen=True, eq=True)
class FunctionDescriptor:
    # COMPAT: Replace with slots=True when we drop Python 3.9
    __slots__ = (
        "alias",
        "exported",
        "findable",
        "internal",
        "method_name",
        "module",
        "name",
        "namespace",
        "output",
        "path",
        "qualname",
    )

    name: str
    namespace: str
    path: str
    exported: bool
    internal: bool
    findable: bool
    alias: bool
    output: str
    method_name: str
    module: str
    qualname: str

    @property
    def cls(self) -> type[Plugin]:
        return load(self)

    @property
    def func(self) -> Callable[..., Any]:
        return getattr(self.cls, self.method_name)

    @property
    def record(self) -> RecordDescriptor | list[RecordDescriptor] | None:
        return getattr(self.func, "__record__", None)

    @property
    def args(self) -> list[tuple[list[str], dict[str, Any]]]:
        return getattr(self.func, "__args__", [])


@dataclass(frozen=True, eq=True)
class FailureDescriptor:
    # COMPAT: Replace with slots=True when we drop Python 3.9
    __slots__ = ("module", "stacktrace")

    module: str
    stacktrace: list[str]


# COMPAT: Add slots=True when we drop Python 3.9
# We can't manually define __slots__ here because we use have to use field() for the default_factory
@dataclass(frozen=True)
class PluginDescriptorLookup:
    # All regular plugins
    # {"<module_path>": PluginDescriptor}
    __regular__: dict[str, PluginDescriptor] = field(default_factory=dict)
    # All OS plugins
    # {"<module_path>": PluginDescriptor}
    __os__: dict[str, PluginDescriptor] = field(default_factory=dict)
    # All child plugins
    # {"<module_path>": PluginDescriptor}
    __child__: dict[str, PluginDescriptor] = field(default_factory=dict)


# COMPAT: Add slots=True when we drop Python 3.9
# We can't manually define __slots__ here because we use have to use field() for the default_factory
@dataclass(frozen=True)
class FunctionDescriptorLookup:
    # All regular plugins
    # {"<function_name>": {"<module_path>": FunctionDescriptor}}
    __regular__: dict[str, dict[str, FunctionDescriptor]] = field(default_factory=dict)
    # All OS plugins
    # {"<function_name>": {"<module_path>": FunctionDescriptor}}
    __os__: dict[str, dict[str, FunctionDescriptor]] = field(default_factory=dict)
    # All child plugins
    # {"<function_name>": {"<module_path>": FunctionDescriptor}}
    __child__: dict[str, dict[str, FunctionDescriptor]] = field(default_factory=dict)


_OSTree: TypeAlias = dict[str, "_OSTree"]


# COMPAT: Add slots=True when we drop Python 3.9
# We can't manually define __slots__ here because we use have to use field() for the default_factory
@dataclass(frozen=True)
class PluginRegistry:
    # Plugin descriptor lookup
    __plugins__: PluginDescriptorLookup = field(default_factory=PluginDescriptorLookup)
    # Function descriptor lookup
    __functions__: FunctionDescriptorLookup = field(default_factory=FunctionDescriptorLookup)
    # OS plugin tree
    # {"<module_part>": {"<module_part>": {}}}
    __ostree__: _OSTree = field(default_factory=dict)
    # Failures
    __failed__: list[FailureDescriptor] = field(default_factory=list)


PLUGINS: PluginRegistry = PluginRegistry()
"""The plugin registry.

Note: It's very important that all values in this dictionary are serializable.
The plugin registry can be stored in a file and loaded later. Plain Python syntax is used to store the registry.
An exception is made for :class:`FailureDescriptor`, :class:`FunctionDescriptor` and :class:`PluginDescriptor`.
"""

GENERATED = False


def export(*args, **kwargs) -> Callable[..., Any]:
    """Decorator to be used on Plugin functions that should be exported.

    Supported keyword arguments:
        property (bool): Whether this export should be regarded as a property.
            Properties are implicitly cached.

        cache (bool): Whether the result of this function should be cached.

        record (RecordDescriptor): The :class:`flow.record.RecordDescriptor` for the records that this function yields.
            If multiple record types are yielded, specificy each descriptor in a list or tuple.
            If the records are dynamically made, use :func:`dissect.target.helpers.record.DynamicDescriptor` instead.

        output (str): The output type of this function. Must be one of:
        - default: Single return value
        - record: Yields records. Implicit when record argument is given.
        - yield: Yields printable values.
        - none: No return value. Plugin is responsible for output formatting and should return ``None``.

    The ``export`` decorator adds some additional private attributes to an exported method or property:
    - ``__output__``: The output type to expect for this function, this is the same as ``output``.
    - ``__record__``: The type of record to expect, this value is the same as ``record``.
    - ``__exported__``: set to ``True`` to indicate the method or property is exported.

    Raises:
        ValueError: if there was an invalid output type.

    Returns:
        An exported function from a plugin.
    """

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any] | property:
        # Properties are implicitly cached
        # Important! Currently it's crucial that this is *always* called
        # See the comment in Plugin.__init_subclass__ for more detail regarding Plugin.__call__
        obj = cache.wrap(obj, no_cache=not kwargs.get("cache", True), cls=kwargs.get("cls"))

        output = kwargs.get("output", "default")
        if output not in OUTPUTS:
            options = ", ".join(OUTPUTS)
            raise ValueError(f'Invalid output method "{output}", must be one of {options}')

        record = kwargs.get("record")
        if record is not None:
            output = "record"

            if not isinstance(record, (list, RecordDescriptor)) and not (
                isinstance(record, (list, tuple)) and all(isinstance(r, RecordDescriptor) for r in record)
            ):
                raise TypeError("record must be an instance of RecordDescriptor or a list of RecordDescriptor")

        obj.__output__ = output
        obj.__record__ = list(record) if isinstance(record, tuple) else record
        obj.__exported__ = True

        if kwargs.get("property", False):
            obj = property(obj)

        return obj

    if len(args) == 1:
        return decorator(args[0])
    return decorator


def internal(*args, **kwargs) -> Callable[..., Any]:
    """Decorator to be used on plugin functions that should be internal only.

    Making a plugin internal means that it's only callable from the Python API and not through ``target-query``.

    This decorator adds the ``__internal__`` private attribute to a method or property.
    The attribute is always set to ``True``, to tell :func:`register` that it is an internal
    method or property.
    """

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any] | property:
        obj.__internal__ = True
        if kwargs.get("property", False):
            obj = property(obj)
        return obj

    if len(args) == 1:
        return decorator(args[0])
    return decorator


def arg(*args, **kwargs) -> Callable[..., Any]:
    """Decorator to be used on Plugin functions that accept additional command line arguments.

    Command line arguments can be added using the ``@arg`` decorator.
    Arguments to this decorator are directly forwarded to the ``ArgumentParser.add_argument`` function of ``argparse``.
    Resulting arguments are passed to the function using kwargs.
    The keyword argument name must match the argparse argument name.

    This decorator adds the ``__args__`` private attribute to a method or property.
    This attribute holds all the command line arguments that were added to the plugin function.
    """

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any]:
        if not hasattr(obj, "__args__"):
            obj.__args__ = []

        obj.__args__.insert(0, (args, kwargs))

        return obj

    return decorator


def alias(*args, **kwargs: dict[str, Any]) -> Callable[..., Any]:
    """Decorator to be used on :class:`Plugin` functions to register an alias of that function."""

    if not kwargs.get("name") and not args:
        raise ValueError("Missing argument 'name'")

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any]:
        if not hasattr(obj, "__aliases__"):
            obj.__aliases__ = []

        if name := (kwargs.get("name") or args[0]):
            obj.__aliases__.append(name)

        return obj

    return decorator


def clone_alias(cls: type, attr: Callable[..., Any], alias: str) -> None:
    """Clone the given attribute to an alias in the provided class."""

    # Clone the function object
    clone = type(attr)(attr.__code__, attr.__globals__, alias, attr.__defaults__, attr.__closure__)
    clone.__kwdefaults__ = attr.__kwdefaults__

    # Copy some attributes
    functools.update_wrapper(clone, attr)
    if wrapped := getattr(attr, "__wrapped__", None):
        # update_wrapper sets a new wrapper, we want the original
        clone.__wrapped__ = wrapped

    # Update module path so we can fool inspect.getmodule with subclassed Plugin classes
    clone.__module__ = cls.__module__

    # Update the names
    clone.__name__ = alias
    clone.__qualname__ = f"{cls.__name__}.{alias}"

    setattr(cls, alias, clone)


class Plugin:
    """Base class for plugins.

    Plugins can optionally be namespaced by specifying the ``__namespace__``
    class attribute. Namespacing results in your plugin needing to be prefixed
    with this namespace when being called. For example, if your plugin has
    specified ``test`` as namespace and a function called ``example``, you must
    call your plugin with ``test.example``.

    A ``Plugin`` class has the following private class attributes:

    - ``__namespace__``
    - ``__record_descriptors__``

    With the following two being assigned in :func:`register`:

    - ``__functions__``
    - ``__exports__``

    Additionally, the methods and attributes of :class:`Plugin` receive more private attributes
    by using decorators.

    The :func:`export` decorator adds the following private attributes

    - ``__exported__``
    - ``__output__``: Set with the :func:`export` decorator.
    - ``__record__``: Set with the :func:`export` decorator.

    The :func:`internal` decorator and :class:`InternalPlugin` set the ``__internal__`` attribute.
    Finally. :func:`args` decorator sets the ``__args__`` attribute.

    The :func:`alias` decorator populates the ``__aliases__`` private attribute of :class:`Plugin` methods.
    Resulting clones of the :class:`Plugin` are populated with the boolean ``__alias__`` attribute set to ``True``.

    Args:
        target: The :class:`~dissect.target.target.Target` object to load the plugin for.
    """

    __namespace__: str = None
    """Defines the plugin namespace."""
    __record_descriptors__: list[RecordDescriptor] = None
    """Defines a list of :class:`~flow.record.RecordDescriptor` of the exported plugin functions."""
    __register__: bool = True
    """Determines whether this plugin will be registered."""
    __findable__: bool = True
    """Determines whether this plugin will be revealed when using search patterns.

    Some (meta)-plugins are not very suitable for wild cards on CLI or
    plugin searches, because they will produce duplicate records or results.
    For instance a plugin that offers the same functions as subplugins will
    produce redundant results when used with a wild card
    (browser.* -> browser.history + browser.*.history).
    """
    __functions__: list[str]
    """Internal. A list of all method names decorated with ``@internal`` or ``@export``."""
    __exports__: list[str]
    """Internal. A list of all method names decorated with ``@export``."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        record_descriptors = _get_descriptors_on_nonprivate_methods(cls)
        cls.__record_descriptors__ = record_descriptors
        # This is a bit tricky currently
        # cls.__call__ is the *function* Plugin.__call__, not from the subclass
        # export() currently will _always_ return a new object because it always calls ``cache.wrap(obj)``
        # This allows this to work, otherwise the Plugin.__call__ would get all the plugin attributes set on it
        cls.__call__ = export(output="record", record=record_descriptors, cache=False, cls=cls)(cls.__call__)

        # Do not register the "base" subclasses defined in this file
        if cls.__module__ != Plugin.__module__:
            register(cls)

    def __init__(self, target: Target):
        self.target = target

    def __getattr__(self, name: str, /) -> Any:
        # "Magic" attribute access is only allowed on namespace plugins, to allow for indirect nesting of namespaces
        # E.g. `__namespace__ = "foo.bar"`, we should try to resolve `bar` on a plugin instance of `foo`
        if not self.__namespace__:
            raise AttributeError(name)

        # Attempt to lookup the attribute using the namespace
        try:
            _, func = self.target.get_function(f"{self.__class__.__namespace__}.{name}")
        except PluginNotFoundError:
            raise AttributeError(name)

        return func

    def is_compatible(self) -> bool:
        """Perform a compatibility check with the target."""
        try:
            self.check_compatible()
        except UnsupportedPluginError:
            return False

        return True

    def check_compatible(self) -> None:
        """Perform a compatibility check with the target.

        This function should return ``None`` if the plugin is compatible with the current target (``self.target``).
        For example, check if a certain file exists. Otherwise it should raise an
        :class:`UnsupportedPluginError`.

        Raises:
            UnsupportedPluginError: If the plugin could not be loaded.
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs) -> Iterator[Record | Any]:
        """Return the records of all exported methods.

        Raises:
            PluginError: If the subclass is not a namespace plugin.
        """
        if not self.__namespace__:
            raise PluginError(f"Plugin {self.__class__.__name__} is not a callable")

        for method_name in self.__exports__:
            if method_name == "__call__":
                continue

            method = getattr(self, method_name)
            if getattr(method, "__output__", None) not in ("record", "yield"):
                continue

            try:
                yield from method()
            except Exception as e:
                self.target.log.error("Error while executing `%s.%s`: %s", self.__namespace__, method_name, e)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

    def get_paths(self) -> Iterator[Path]:
        if self.target.is_direct:
            yield from self._get_paths_direct()
        else:
            yield from self._get_paths()

    def _get_paths_direct(self) -> Iterator[Path]:
        """Return all paths as given by the user."""
        for path in self.target._loader.paths:
            yield self.target.fs.path(str(path))

    def _get_paths(self) -> Iterator[Path]:
        """Return all files of interest to the plugin.

        To be implemented by the plugin subclass.
        """
        raise NotImplementedError


def register(plugincls: type[Plugin]) -> None:
    """Register a plugin, and put related data inside :attr:`PLUGINS`.

    This function uses the following private attributes that are set using decorators:
        - ``__exported__``: Set in :func:`export`.
        - ``__internal__``: Set in :func:`internal`.

    Additionally, ``register`` sets the following private attributes on the `plugincls`:
        - ``__functions__``: A list of all the methods and properties that are ``__internal__`` or ``__exported__``.
        - ``__exports__``: A list of all the methods or properties that were explicitly exported.

    If a plugincls ``__register__`` attribute is set to ``False``, the plugin will not be registered, but the
    plugin will still be processed for the private attributes mentioned above.

    Args:
        plugincls: A plugin class to register.

    Raises:
        ValueError: If ``plugincls`` is not a subclass of :class:`Plugin`.
    """
    if not issubclass(plugincls, Plugin):
        raise TypeError("Not a subclass of Plugin")

    # Register the plugin in the correct tree
    __plugins__ = PLUGINS.__plugins__
    __functions__ = PLUGINS.__functions__
    if issubclass(plugincls, OSPlugin):
        plugin_index = __plugins__.__os__
        function_index = __functions__.__os__
    elif issubclass(plugincls, ChildTargetPlugin):
        plugin_index = __plugins__.__child__
        function_index = __functions__.__child__
    else:
        plugin_index = __plugins__.__regular__
        function_index = __functions__.__regular__

    exports = []
    functions = []
    module_path = _module_path(plugincls)

    # This enables plugin directories, e.g.:
    # <plugin>/_plugin.py
    # <plugin>/helpers.py
    module_path = module_path.removesuffix("._plugin")

    module_key = f"{module_path}.{plugincls.__qualname__}"

    if not issubclass(plugincls, ChildTargetPlugin):
        # First pass to resolve aliases
        for attr in _get_nonprivate_attributes(plugincls):
            for alias in getattr(attr, "__aliases__", []):
                clone_alias(plugincls, attr, alias)

        for attr in _get_nonprivate_attributes(plugincls):
            if isinstance(attr, property):
                attr = attr.fget

            if getattr(attr, "__generated__", False) and plugincls != plugincls.__nsplugin__:
                continue

            exported = getattr(attr, "__exported__", False)
            internal = getattr(attr, "__internal__", False)

            if exported or internal:
                functions.append(attr.__name__)
                if exported:
                    exports.append(attr.__name__)

                if plugincls.__register__:
                    name = attr.__name__
                    if plugincls.__namespace__:
                        name = f"{plugincls.__namespace__}.{name}"

                    path = f"{module_path}.{attr.__name__}"

                    descriptor = FunctionDescriptor(
                        name=name,
                        namespace=plugincls.__namespace__,
                        path=path,
                        exported=exported,
                        internal=internal,
                        findable=plugincls.__findable__,
                        alias=name in getattr(attr, "__aliases__", []),
                        output=getattr(attr, "__output__", "none"),
                        method_name=attr.__name__,
                        module=plugincls.__module__,
                        qualname=plugincls.__qualname__,
                    )

                    # Register the functions in the lookup
                    function_index.setdefault(name, {})[module_key] = descriptor

    if plugincls.__namespace__:
        # Namespaces are also callable, so register the namespace itself as well
        # NamespacePlugin needs to register itself for every additional subclass, so allow overwrites here
        functions.append("__call__")
        if exports:
            exports.append("__call__")

        if plugincls.__register__:
            descriptor = FunctionDescriptor(
                name=plugincls.__namespace__,
                namespace=plugincls.__namespace__,
                path=module_path,
                exported=len(exports) != 0,
                internal=len(functions) != 0 and len(exports) == 0,
                findable=plugincls.__findable__,
                alias=False,
                output=getattr(plugincls.__call__, "__output__", "none"),
                method_name="__call__",
                module=plugincls.__module__,
                qualname=plugincls.__qualname__,
            )

            function_index.setdefault(plugincls.__namespace__, {})[module_key] = descriptor

    # Update the class with the plugin attributes
    plugincls.__functions__ = functions
    plugincls.__exports__ = exports

    if plugincls.__register__:
        plugin_index[module_key] = PluginDescriptor(
            module=plugincls.__module__,
            qualname=plugincls.__qualname__,
            namespace=plugincls.__namespace__,
            path=module_path,
            findable=plugincls.__findable__,
            functions=functions,
            exports=exports,
        )

        if issubclass(plugincls, OSPlugin):
            # Also store the OS plugins in a tree by module path
            # This is used to filter plugins based on the OSPlugin subclass
            # We don't store anything at the end of the tree, as we only use the tree to check if a plugin is compatible

            # Also slightly modify the module key to allow for more efficient filtering later
            # This is done by removing the last two parts of the module key, which are the file name and the class name
            module_parts = module_key.split(".")
            if module_parts[-2] != "_os":
                log.warning("OS plugin modules should be named as <os>/_os.py: %s", module_key)

            obj = PLUGINS.__ostree__
            for part in module_parts[:-2]:
                obj = obj.setdefault(part, {})

        log.debug("Plugin registered: %s", module_key)


def _get_plugins() -> PluginRegistry:
    """Load the plugin registry, or generate it if it doesn't exist yet."""
    global PLUGINS, GENERATED

    if not GENERATED:
        try:
            from dissect.target.plugins._pluginlist import PLUGINS
        except ImportError:
            PLUGINS = generate()

        GENERATED = True

    return PLUGINS


def _module_path(cls: type[Plugin] | str) -> str:
    """Returns the module path relative to ``dissect.target.plugins``."""
    if issubclass(cls, Plugin):
        module = getattr(cls, "__module__", "")
    elif isinstance(cls, str):
        module = cls
    else:
        raise TypeError(f"Invalid argument type: {cls}")

    return module.replace(MODULE_PATH, "").lstrip(".")


@functools.lru_cache(256)
def _os_match(osfilter: type[OSPlugin], module_path: str) -> bool:
    """Check if the a plugin is compatible with the given OS filter."""
    module_parts = module_path.split(".")

    # DefaultOSPlugin is "compatible" with everything, even other `os.*` modules (except `_os`)
    if issubclass(osfilter, default._os.DefaultOSPlugin) and (len(module_parts) < 2 or module_parts[-2] != "_os"):
        return True

    os_parts = _module_path(osfilter).split(".")[:-1]

    obj = _get_plugins().__ostree__
    for plugin_part, os_part in zip_longest(module_parts, os_parts):
        if plugin_part not in obj:
            break

        if plugin_part != os_part:
            return False

        obj = obj[plugin_part]

    return True


def plugins(osfilter: type[OSPlugin] | None = None, *, index: str = "__regular__") -> Iterator[PluginDescriptor]:
    """Walk the plugin registry and return plugin descriptors.

    If ``osfilter`` is specified, only plugins related to the provided OSPlugin, or plugins
    with no OS relation are returned. If ``osfilter`` is ``None``, all plugins will be returned.

    One exception to this is if the ``osfilter`` is a (sub-)class of ``DefaultOSPlugin``, then plugins
    are returned as if no ``osfilter`` was specified.

    The ``index`` parameter can be used to specify the index to return plugins from. By default,
    this is set to return regular plugins. Other possible values are ``__os__`` and ``__child__``.
    These return :class:`OSPlugin` and :class:`ChildTargetPlugin` respectively.

    Args:
        osfilter: The optional :class:`OSPlugin` to filter the returned plugins on.
        index: The plugin index to return plugins from. Defaults to regular plugins.

    Yields:
        Plugin descriptors in the plugin registry based on the given filter criteria.
    """

    plugin_index: dict[str, PluginDescriptor] = getattr(_get_plugins().__plugins__, index, {})

    # This is implemented as a list comprehension for performance reasons!
    yield from (
        descriptor
        for module_path, descriptor in plugin_index.items()
        if (index != "__os__" and (osfilter is None or _os_match(osfilter, module_path)))
        or (index == "__os__" and (osfilter is None or osfilter.__module__ == descriptor.module))
    )


def os_plugins() -> Iterator[PluginDescriptor]:
    """Retrieve all OS plugin descriptors."""
    yield from plugins(index="__os__")


def child_plugins() -> Iterator[PluginDescriptor]:
    """Retrieve all child plugin descriptors."""
    yield from plugins(index="__child__")


def functions(osfilter: type[OSPlugin] | None = None, *, index: str = "__regular__") -> Iterator[FunctionDescriptor]:
    """Retrieve all function descriptors.

    Args:
        osfilter: The optional :class:`OSPlugin` to filter the returned functions on.
        index: The plugin index to return functions from. Defaults to regular functions.

    Yields:
        Function descriptors in the plugin registry based on the given filter criteria.
    """

    function_index: dict[str, dict[str, FunctionDescriptor]] = getattr(_get_plugins().__functions__, index, {})

    # This is implemented as a list comprehension for performance reasons!
    yield from (
        descriptor
        for function_index in function_index.values()
        for module_path, descriptor in function_index.items()
        if osfilter is None or _os_match(osfilter, module_path)
    )


def lookup(
    function_name: str, osfilter: type[OSPlugin] | None = None, *, index: str = "__regular__"
) -> Iterator[FunctionDescriptor]:
    """Lookup a function descriptor by function name.

    Args:
        func_name: Function name to lookup.
        osfilter: The optional ``OSPlugin`` to filter results with for compatibility.
        index: The plugin index to return plugins from. Defaults to regular functions.

    Yields:
        Function descriptors that match the given function name and filter criteria.
    """

    function_index: dict[str, FunctionDescriptor] = getattr(_get_plugins().__functions__, index, {}).get(
        function_name, {}
    )

    # This is implemented as a list comprehension for performance reasons!
    entries: Iterator[FunctionDescriptor] = (
        value for key, value in function_index.items() if osfilter is None or _os_match(osfilter, key)
    )

    yield from sorted(entries, key=lambda desc: desc.module.count("."), reverse=True)


def load(desc: FunctionDescriptor | PluginDescriptor) -> type[Plugin]:
    """Helper function that loads a plugin from a given function or plugin descriptor.

    Args:
        desc: Function descriptor as returned by :func:`plugin.lookup` or plugin descriptor
              as returned by :func:`plugin.plugins`.

    Returns:
        The plugin class.

    Raises:
        PluginError: Raised when any other exception occurs while trying to load the plugin.
    """
    module = desc.module

    try:
        obj = importlib.import_module(module)
        for part in desc.qualname.split("."):
            obj = getattr(obj, part)
    except Exception as e:
        raise PluginError(f"An exception occurred while trying to load a plugin: {module}") from e
    else:
        return obj


def os_match(target: Target, descriptor: PluginDescriptor) -> bool:
    """Check if a plugin descriptor is compatible with the target OS.

    Args:
        target: The target to check compatibility with.
        descriptor: The plugin descriptor to check compatibility for.
    """
    return _os_match(target._os_plugin, f"{descriptor.module}.{descriptor.qualname}")


def failed() -> list[FailureDescriptor]:
    """Return all plugins that failed to load."""
    return _get_plugins().__failed__


@functools.cache
def _generate_long_paths(os_filter: type[OSPlugin]) -> dict[str, list[FunctionDescriptor]]:
    """Generate a dictionary of all long paths to their function descriptors for the given functions."""
    __regular__ = functions(os_filter, index="__regular__")
    __os__ = functions(os_filter, index="__os__")

    paths = {}
    for descriptor in chain(__regular__, __os__):
        # Namespace plugins are callable so exclude the explicit __call__ method
        if descriptor.method_name == "__call__":
            continue

        paths.setdefault(descriptor.path, []).append(descriptor)

    return paths


def find_functions(
    patterns: str,
    target: Target | None = None,
    compatibility: bool = False,
    ignore_load_errors: bool = False,
    show_hidden: bool = False,
) -> tuple[list[FunctionDescriptor], set[str]]:
    """Finds exported plugin functions that match the target and the patterns.

    Given a target, a comma separated list of patterns and an optional compatibility flag,
    this function finds matching plugins, optionally checking compatibility.

    Returns:
        A tuple containing a list of matching function descriptors and a set of invalid patterns.
    """
    found = []

    __functions__ = _get_plugins().__functions__

    regular_functions = __functions__.__regular__
    os_functions = __functions__.__os__

    os_filter = target._os_plugin if target is not None else None

    invalid_functions = set()

    for pattern in patterns.split(","):
        if not pattern:
            continue

        exact_match = pattern in regular_functions
        exact_os_match = pattern in os_functions

        if exact_match or exact_os_match:
            matches = list(_filter_exact_match(pattern, os_filter, "__os__" if exact_os_match else "__regular__"))
        else:
            # If we don't have an exact function match, do a slower treematch
            matches = list(_filter_tree_match(pattern, os_filter, show_hidden=show_hidden))

        if matches:
            found.extend(matches)
        else:
            invalid_functions.add(pattern)

    result = _filter_compatible(found, target, ignore_load_errors) if compatibility and target is not None else found
    return list(result), invalid_functions


def find_functions_by_record_field_type(
    field_types: str | list[str],
    target: Target | None = None,
    compatibility: bool = False,
    ignore_load_errors: bool = False,
) -> Iterator[FunctionDescriptor]:
    """Find functions that yield records with a specific field type.

    Args:
        field_types: The field type to search for.
        target: The target to check compatibility with.
        compatibility: Whether to check compatibility with the target.
        ignore_load_errors: Whether to ignore load errors.
    """

    def _search(search_types: list[str]) -> Iterator[FunctionDescriptor]:
        for descriptor in functions():
            try:
                if descriptor.output != "record" or (record_descriptors := descriptor.record) is None:
                    continue
            except Exception:
                if ignore_load_errors:
                    continue
                raise

            for rd in record_descriptors if isinstance(record_descriptors, list) else [record_descriptors]:
                if any(type_name in search_types for type_name, _ in rd._field_tuples):
                    yield descriptor
                    break

    search_types = field_types if isinstance(field_types, list) else [field_types]
    if compatibility and target is not None:
        yield from _filter_compatible(_search(search_types), target, ignore_load_errors)
    else:
        yield from _search(search_types)


def _filter_exact_match(
    pattern: str,
    os_filter: type[OSPlugin] | None,
    index: str = "__regular__",
) -> Iterator[FunctionDescriptor]:
    descriptors = lookup(pattern, os_filter, index=index)
    yield from (descriptor for descriptor in descriptors if descriptor.exported)


def _filter_tree_match(
    pattern: str, os_filter: type[OSPlugin] | None, show_hidden: bool = False
) -> Iterator[FunctionDescriptor]:
    """Perform a slow tree match on all ``__regular__`` and ``__os__`` :class:`FunctionDescriptor` instances."""
    path_lookup = _generate_long_paths(os_filter)

    # Change the treematch pattern into an fnmatch-able pattern to give back all functions from the sub-tree
    # (if there is a subtree).
    #
    # Examples:
    #     -f apps.webservers.iis -> apps.webservers.iis* (logs etc)
    #     -f apps.webservers.iis.logs -> apps.webservers.iis.logs* (only the logs, there is no subtree)
    # We do not include a dot because that does not work if the full path is given:
    #     -f apps.webservers.iis.logs != apps.webservers.iis.logs.* (does not work)
    search_pattern = pattern
    if not has_glob_magic(pattern):
        search_pattern += "*"

    result: list[FunctionDescriptor] = []
    for path in fnmatch.filter(path_lookup.keys(), search_pattern):
        for descriptor in path_lookup[path]:
            # Skip plugins that don't want to be found by wildcards
            if not descriptor.exported or (not show_hidden and not descriptor.findable):
                continue

            result.append(descriptor)

    yield from sorted(result, key=lambda desc: desc.module.count("."), reverse=True)


def _filter_compatible(
    descriptors: list[FunctionDescriptor], target: Target, ignore_load_errors: bool = False
) -> Iterator[FunctionDescriptor]:
    """Filter a list of function descriptors based on compatibility with a target."""
    compatible = set()
    incompatible = set()

    for descriptor in descriptors:
        if descriptor.qualname in compatible:
            yield descriptor
            continue

        if descriptor.qualname in incompatible:
            continue

        try:
            plugincls = load(descriptor)
        except Exception:
            if ignore_load_errors:
                incompatible.add(descriptor.qualname)
                continue
            raise

        try:
            if issubclass(plugincls, OSPlugin):
                if issubclass(target._os_plugin, plugincls):
                    compatible.add(descriptor.qualname)
                    yield descriptor
                else:
                    incompatible.add(descriptor.qualname)
                    continue
            elif plugincls(target).is_compatible():
                compatible.add(descriptor.qualname)
                yield descriptor
            else:
                incompatible.add(descriptor.qualname)
        except Exception:
            incompatible.add(descriptor.qualname)


def generate() -> dict[str, Any]:
    """Internal function to generate the list of available plugins.

    Walks the plugins directory and imports any ``.py`` files in there.
    Plugins will be automatically registered.

    Returns:
        The global ``PLUGINS`` dictionary.
    """
    plugins_dir = Path(__file__).parent / "plugins"
    for path in _find_py_files(plugins_dir):
        relative_path = path.relative_to(plugins_dir)
        module_tuple = (MODULE_PATH, *relative_path.parent.parts, relative_path.stem)
        load_module_from_name(".".join(module_tuple))

    return PLUGINS


def _find_py_files(path: Path) -> Iterator[Path]:
    """Walk all the files and directories in ``path`` and return all files ending in ``.py``.

    Do not walk or yield paths containing the following names:

    - __pycache__
    - __init__

    Furthermore, it logs an error if ``path`` does not exist.

    Args:
        path: The path to a directory or file to walk and filter.
    """
    if not path.exists():
        log.error("Path %s does not exist", path)
        return

    it = [path] if path.is_file() else path.glob("**/*.py")

    for entry in it:
        if not entry.is_file() or entry.name == "__init__.py":
            continue

        yield entry


def load_module_from_name(module_path: str) -> None:
    """Load a module from ``module_path``."""
    try:
        # This will trigger the __init__subclass__() of the Plugin subclasses in the module.
        importlib.import_module(module_path)
    except Exception as e:
        log.info("Unable to import %s", module_path)
        log.debug("Error while trying to import module %s", module_path, exc_info=e)
        _save_plugin_import_failure(module_path)


def load_module_from_file(path: Path, base_path: Path) -> None:
    """Loads a module from a file indicated by ``path`` relative to ``base_path``.

    The module is added to ``sys.modules`` so it can be found everywhere.

    Args:
        path: The file to load as module.
        base_path: The base directory of the module.
    """
    try:
        relative_path = path.relative_to(base_path)
        module_tuple = (*relative_path.parent.parts, relative_path.stem)
        spec = importlib.util.spec_from_file_location(".".join(module_tuple), path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[module.__name__] = module
    except Exception as e:
        log.error("Unable to import %s", path)  # noqa: TRY400
        log.debug("Error while trying to import module %s", path, exc_info=e)
        _save_plugin_import_failure(str(path))


def load_modules_from_paths(paths: list[Path]) -> None:
    """Iterate over the ``paths`` and load all ``.py`` files."""
    for path in paths:
        for file in _find_py_files(path):
            base_path = path.parent if file == path else path
            load_module_from_file(file, base_path)


def get_external_module_paths(path_list: list[Path]) -> list[Path]:
    """Return a list of external plugin directories."""
    output_list = environment_variable_paths() + path_list

    return list(set(output_list))


def environment_variable_paths() -> list[Path]:
    """Return additional plugin directories specified by the ``DISSECT_PLUGINS`` environment variable."""
    env_var = os.environ.get("DISSECT_PLUGINS")

    plugin_dirs = env_var.split(":") if env_var else []

    return [Path(directory) for directory in plugin_dirs]


def _save_plugin_import_failure(module: str) -> None:
    """Store errors that occurred during plugin import."""
    stacktrace = traceback.format_exception(*sys.exc_info())
    PLUGINS.__failed__.append(FailureDescriptor(module, stacktrace))


def _get_nonprivate_attribute_names(cls: type[Plugin]) -> list[str]:
    """Retrieve all attributes that do not start with ``_``."""
    return [attr for attr in dir(cls) if not attr.startswith("_")]


def _get_nonprivate_attributes(cls: type[Plugin]) -> list[Any]:
    """Retrieve all public attributes of a :class:`Plugin`."""
    # Note: `dir()` might return attributes from parent class
    return [getattr(cls, attr) for attr in _get_nonprivate_attribute_names(cls)]


def _get_nonprivate_methods(cls: type[Plugin]) -> list[Callable]:
    """Retrieve all public methods of a :class:`Plugin`."""
    return [attr for attr in _get_nonprivate_attributes(cls) if not isinstance(attr, property) and callable(attr)]


def _get_descriptors_on_nonprivate_methods(cls: type[Plugin]) -> list[RecordDescriptor]:
    """Return record descriptors set on nonprivate methods in ``cls`` class."""
    descriptors = set()

    for m in _get_nonprivate_methods(cls):
        descriptors.update(_get_descriptors_on_method(m))

    return list(descriptors)


def _get_descriptors_on_method(method: Callable) -> list[RecordDescriptor]:
    """Return record descriptors set on a method."""
    if not (record := getattr(method, "__record__", None)):
        return []

    try:
        # check if __record__ value is iterable (for example, a list)
        return list(record)
    except TypeError:
        return [record]


# Classes for specific types of plugins
# These need to be at the bottom of the module because __init_subclass__ requires everything
# in the parent class Plugin to be defined and resolved.
class OSPlugin(Plugin):
    """Base class for OS plugins.

    This provides a base class for certain common functions of OS's, which each OS plugin has to implement separately.

    For example, it provides an interface for retrieving the hostname and users of a target.

    All derived classes MUST implement ALL the classmethods and exported
    methods with the same ``@classmethod`` or ``@export(...)`` annotation.
    """

    def __init_subclass__(cls, **kwargs):
        # Note that cls is the subclass
        super().__init_subclass__(**kwargs)

        for os_method in _get_nonprivate_attributes(OSPlugin):
            if isinstance(os_method, property):
                os_method = os_method.fget
            os_docstring = os_method.__doc__

            method = getattr(cls, os_method.__name__, None)
            if isinstance(method, property):
                method = method.fget
            # This works as None has a __doc__ property (which is None).
            docstring = method.__doc__

            if method and not docstring:
                if hasattr(method, "__func__"):
                    method = method.__func__
                method.__doc__ = os_docstring

    def check_compatible(self) -> bool:
        """OSPlugin's use a different compatibility check, override the one from the :class:`Plugin` class.

        Returns:
            This function always returns ``True``.
        """
        return True

    @classmethod
    def detect(cls, fs: Filesystem) -> Filesystem | None:
        """Provide detection of this OSPlugin on a given filesystem.

        Args:
            fs: :class:`~dissect.target.filesystem.Filesystem` to detect the OS on.

        Returns:
            The root filesystem / sysvol when found.
        """
        raise NotImplementedError

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        """Initiate this OSPlugin with the given target and detected filesystem.

        Args:
            target: The :class:`~dissect.target.target.Target` object.
            sysvol: The filesystem that was detected in the ``detect()`` function.

        Returns:
            An instantiated version of the OSPlugin.
        """
        raise NotImplementedError

    @export(property=True)
    def hostname(self) -> str | None:
        """Return the target's hostname.

        Returns:
            The hostname as string.
        """
        raise NotImplementedError

    @export(property=True)
    def ips(self) -> list[str]:
        """Return the IP addresses configured in the target.

        Returns:
            The IPs as list.
        """
        raise NotImplementedError

    @export(property=True)
    def version(self) -> str | None:
        """Return the target's OS version.

        Returns:
            The OS version as string.
        """
        raise NotImplementedError

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        """Return the users available in the target.

        Returns:
            A list of user records.
        """
        raise NotImplementedError

    @export(property=True)
    def os(self) -> str:
        """Return a slug of the target's OS name.

        Returns:
            A slug of the OS name, e.g. 'windows' or 'linux'.
        """
        raise NotImplementedError

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        """Yields miscellaneous user paths and user keys.

        Example:
            ("c:/Windows/ServiceProfiles/LocalService", ("sid", "S-1-5-19"))
        """

        # Consider moving this to the UsersPlugin class, and create separate UsersPlugin subclasses for each OS
        # when the machinery concerning users and user_paths becomes more complex.
        raise NotImplementedError

    @internal
    def os_tree(self) -> list[str]:
        """Returns the :func:`os` value of this and all the OS plugin parents."""
        result: list[str] = []
        for klass in self.__class__.mro():
            if not issubclass(klass, OSPlugin):
                # In case the plugin extends multiple different classes.
                continue

            if klass is OSPlugin:
                break

            result.append(klass.os.__get__(self))
        # dicts maintain insertion order, we use it here to get rid of duplicates
        return list(dict.fromkeys(result))

    @export(property=True)
    def architecture(self) -> str | None:
        """Return a slug of the target's OS architecture.

        Returns:
            A slug of the OS architecture, e.g. 'x86_32-unix', 'MIPS-linux' or
            'AMD64-win32', or 'unknown' if the architecture is unknown.
        """
        raise NotImplementedError


class ChildTargetPlugin(Plugin):
    """A Child target is a special plugin that can list more Targets.

    For example, :class:`~dissect.target.plugins.child.esxi.ESXiChildTargetPlugin` can
    list all of the Virtual Machines on the host.
    """

    __type__ = None

    def list_children(self) -> Iterator[ChildTargetRecord]:
        """Yield :class:`~dissect.target.helpers.record.ChildTargetRecord` records of all
        possible child targets on this target.
        """
        raise NotImplementedError


class NamespacePlugin(Plugin):
    """A namespace plugin provides services to access functionality from a group of subplugins.

    Support is currently limited to shared exported functions with output type ``record`` and ``yield``.
    """

    def __init_subclass__(cls, **kwargs):
        # Upon subclassing, decide whether this is a direct subclass of NamespacePlugin
        # If this is not the case, autogenerate aggregate methods for methods record output.

        # Do not process the "base" subclasses defined in this file
        if cls.__module__ == NamespacePlugin.__module__:
            super().__init_subclass__(**kwargs)
            return

        if cls.__bases__[0] not in (NamespacePlugin, InternalNamespacePlugin):
            # This is a subclass of a subclass of (Internal)NamespacePlugin
            # We need to aggregate the methods of this subclass
            cls.__findable__ = True
            # Run the normal subclass initialization first so the plugin gets all its exports registered
            super().__init_subclass__(**kwargs)
            cls.__init_subclass_subplugin__(cls, **kwargs)
            cls.__nsplugin__.__init_subclass__(**kwargs)
        else:
            # This is a direct subclass of (Internal)NamespacePlugin
            # Not much to do here, just do some initialization and register the plugin
            cls.__findable__ = False
            super().__init_subclass__(**kwargs)

            cls.__nsplugin__ = cls
            if not hasattr(cls, "__subplugins__"):
                cls.__subplugins__ = set()

    def __init_subclass_subplugin__(cls, **kwargs) -> None:
        # Register the current plugin class as a subplugin with
        # the direct subclass of NamespacePlugin
        cls.__nsplugin__.__subplugins__.add(cls.__namespace__)

        # Generate a tuple of class names for which we do not want to add subplugin functions, which is the
        # NamespacePlugin and all of its superclasses (minus the base object)
        reserved_cls = tuple({klass.__name__ for klass in cls.__nsplugin__.mro() if klass is not object})

        # Collect the public exports of the subplugin
        for export_name in cls.__exports__:
            export_attr = getattr(cls, export_name)

            # The export need to be callable and may not be part of a parent class
            if not isinstance(export_attr, Callable) or export_attr.__qualname__.startswith(reserved_cls):
                continue

            # The export needs to output records or yield something else
            if (output_type := export_attr.__output__) not in ["record", "yield"]:
                continue

            if aggregator := getattr(cls.__nsplugin__, export_name, None):
                # If we already have an aggregate method, update some fields
                if not hasattr(aggregator, "__subplugins__"):
                    # This is not an aggregator, but a re-implementation of a subclass method by the subplugin
                    continue

                if aggregator.__output__ != output_type:
                    # The output type of the existing aggregator is different, so we can't merge them
                    raise PluginError(
                        "Cannot merge namespace methods with different output types"
                        f" ({aggregator.__output__} != {output_type} in {cls.__namespace__}.{export_name})"
                    )

                # Update record descriptors
                if (export_attr.__record__, aggregator.__record__) != (None, None):
                    aggregator.__record__ = list(
                        set(_get_descriptors_on_method(export_attr) + _get_descriptors_on_method(aggregator))
                    )
            else:
                # Generate a method for the parent namespace class
                aggregator = export(
                    output=output_type,
                    record=export_attr.__record__,
                    cache=False,
                    cls=cls.__nsplugin__,
                )(cls.__generate_aggregator(cls.__nsplugin__, export_name))

                # Add to the parent namespace class
                setattr(cls.__nsplugin__, export_name, aggregator)

            # Add subplugin to aggregator
            aggregator.__subplugins__.append(cls.__namespace__)
            cls.__update_aggregator_docs(aggregator)

    def check_compatible(self) -> None:
        at_least_one = False
        for entry in self.__subplugins__:
            try:
                self.target.get_function(entry)
                at_least_one = True
            except UnsupportedPluginError:  # noqa: PERF203
                self.target.log.warning("Subplugin %s is not compatible with target", entry)
            except Exception as e:
                self.target.log.error("Failed to load subplugin: %s", entry)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

        if not at_least_one:
            raise UnsupportedPluginError("No compatible subplugins found")

    @staticmethod
    def __generate_aggregator(nsklass: type[NamespacePlugin], method_name: str) -> Callable:
        def aggregator(self: NamespacePlugin) -> Iterator[Record]:
            for ns in aggregator.__subplugins__:
                try:
                    yield from self.target.get_function(f"{ns}.{method_name}")[1]()
                except UnsupportedPluginError:  # noqa: PERF203
                    continue
                except Exception as e:
                    self.target.log.error("Subplugin %s.%s raised an exception: %s", ns, method_name, e)  # noqa: TRY400
                    self.target.log.debug("", exc_info=e)

        # Holds the subplugins that share this method
        aggregator.__subplugins__ = []
        aggregator.__generated__ = True

        # Clone some names and attributes
        aggregator.__module__ = nsklass.__module__
        aggregator.__name__ = method_name
        aggregator.__qualname__ = f"{nsklass.__name__}.{method_name}"

        return aggregator

    @staticmethod
    def __update_aggregator_docs(aggregator: Callable) -> None:
        aggregator.__doc__ = f"Return {aggregator.__name__} for: {', '.join(aggregator.__subplugins__)}"


__INTERNAL_PLUGIN_METHOD_NAMES__ = {attr.__name__ for attr in _get_nonprivate_methods(Plugin)}


class InternalPlugin(Plugin):
    """Parent class for internal plugins.

    InternalPlugin marks all non-private methods internal by default
    (same as ``@internal`` decorator).
    """

    def __init_subclass__(cls, **kwargs):
        for method in _get_nonprivate_methods(cls):
            if callable(method) and method.__name__ not in __INTERNAL_PLUGIN_METHOD_NAMES__:
                method.__internal__ = True

        super().__init_subclass__(**kwargs)
        return cls


class InternalNamespacePlugin(NamespacePlugin, InternalPlugin):
    pass



from __future__ import annotations

import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.ssh.opensshd import SSHServerPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_sshd_config_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    config_file = absolute_path("_data/plugins/apps/ssh/opensshd/sshd_config")
    plugin = SSHServerPlugin(target_unix_users)
    fs_unix.map_file(str(plugin.sshd_config_path), config_file)

    target_unix_users.add_plugin(SSHServerPlugin)
    results = list(target_unix_users.opensshd.config())

    assert len(results) == 1
    result = results[0]

    assert str(result.source) == str(plugin.sshd_config_path)
    assert result.Port == [22]
    assert result.LoginGraceTime == "2m"
    assert result.PermitRootLogin == "prohibit-password"
    assert result.StrictModes
    assert result.MaxAuthTries == 6
    assert not hasattr(result, "PubkeyAuthentication")
    assert result.AuthorizedKeysFile == ".ssh/authorized_keys"


def test_sshd_config_plugin_multiple_definitions(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    config = """
    Port 22
    Port 1234
    ListenAddress 1.2.3.4
    ListenAddress 9.8.7.6
    """

    plugin = SSHServerPlugin(target_unix_users)
    fs_unix.map_file_fh(
        str(plugin.sshd_config_path),
        BytesIO(textwrap.dedent(config).encode()),
    )

    target_unix_users.add_plugin(SSHServerPlugin)
    results = list(target_unix_users.opensshd.config())

    assert len(results) == 1
    result = results[0]

    assert str(result.source) == str(plugin.sshd_config_path)
    assert result.Port == [22, 1234]
    assert result.ListenAddress == ["1.2.3.4", "9.8.7.6"]



from __future__ import annotations

from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch
from zoneinfo import ZoneInfo

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.os.unix.log.auth import AuthPlugin
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import VirtualFilesystem


def test_auth(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))

    data_path = "_data/plugins/os/unix/log/auth/auth.log"
    data_file = absolute_path(data_path)
    fs_unix.map_file("var/log/auth.log", data_file)

    entry = fs_unix.get("var/log/auth.log")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 10
        assert results[-1].ts == dt(2022, 11, 14, 6, 39, 1, tzinfo=ZoneInfo("Europe/Amsterdam"))
        assert results[-1].message == "pam_unix(cron:session): session opened for user root by (uid=0)"


def test_auth_with_gz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Pacific/Honolulu"))

    empty_file = absolute_path("_data/plugins/os/unix/log/empty.log")
    fs_unix.map_file("var/log/auth.log", empty_file)

    gz_path = "_data/plugins/os/unix/log/auth/auth.log.gz"
    gz_file = absolute_path(gz_path)
    fs_unix.map_file("var/log/auth.log.gz", gz_file)

    entry = fs_unix.get("var/log/auth.log.gz")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 10
        assert results[-1].ts == dt(2022, 11, 14, 6, 39, 1, tzinfo=ZoneInfo("Pacific/Honolulu"))
        assert results[-1].message == "pam_unix(cron:session): session opened for user root by (uid=0)"


def test_auth_with_bz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"America/Nuuk"))

    empty_file = absolute_path("_data/plugins/os/unix/log/empty.log")
    fs_unix.map_file("var/log/auth.log", empty_file)

    bz_path = "_data/plugins/os/unix/log/auth/auth.log.bz2"
    bz_file = absolute_path(bz_path)
    fs_unix.map_file("var/log/auth.log.1.bz2", bz_file)

    entry = fs_unix.get("var/log/auth.log.1.bz2")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 10
        assert results[-1].ts == dt(2022, 11, 14, 6, 39, 1, tzinfo=ZoneInfo("America/Nuuk"))
        assert results[-1].message == "pam_unix(cron:session): session opened for user root by (uid=0)"


def test_auth_year_rollover(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Etc/UTC"))

    data_path = "_data/plugins/os/unix/log/auth/secure"
    data_file = absolute_path(data_path)
    fs_unix.map_file("var/log/secure", data_file)

    entry = fs_unix.get("var/log/secure")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 2
        results.reverse()
        assert results[0].ts == dt(2021, 12, 31, 3, 14, 0, tzinfo=ZoneInfo("Etc/UTC"))
        assert results[1].ts == dt(2022, 1, 1, 13, 37, 0, tzinfo=ZoneInfo("Etc/UTC"))


@pytest.mark.parametrize(
    ("message", "results"),
    [
        pytest.param(
            "Mar 29 10:43:01 ubuntu-1 sshd[1193]: Accepted password for test_user from 8.8.8.8 port 52942 ssh2",
            {
                "service": "sshd",
                "pid": 1193,
                "action": "accepted authentication",
                "authentication_type": "password",
                "user": "test_user",
                "remote_ip": "8.8.8.8",
                "port": 52942,
            },
            id="sshd: accepted password",
        ),
        pytest.param(
            "Jun 4 22:14:15 ubuntu-1 sshd[41458]: Failed password for root from 8.8.8.8 port 22 ssh2",
            {
                "service": "sshd",
                "pid": 41458,
                "action": "failed authentication",
                "authentication_type": "password",
                "user": "root",
                "remote_ip": "8.8.8.8",
                "port": 22,
            },
            id="sshd: failed password",
        ),
        pytest.param(
            "Jun 4 22:14:15 ubuntu-1 sshd[12345]: reverse mapping checking getaddrinfo for some-hostname-with-digits-012.34.56.78.example.com [90.12.34.56] failed - POSSIBLE BREAK-IN ATTEMPT!",  # noqa: E501
            {
                "service": "sshd",
                "pid": 12345,
                "remote_ips": ["90.12.34.56", "12.34.56.78"],
            },
            id="sshd: reverse dns ip addr",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: Accepted publickey for test_user "
            "from 8.8.8.8 port 12345 ssh2: RSA SHA256:123456789asdfghjklertzuio",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "accepted authentication",
                "authentication_type": "publickey",
                "user": "test_user",
                "remote_ip": "8.8.8.8",
                "port": 12345,
                "ssh_protocol": "ssh2",
                "encryption_algorithm": "RSA",
                "hash_algorithm": "SHA256",
                "key_hash": "123456789asdfghjklertzuio",
            },
            id="sshd: accepted publickey",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: Failed publickey for test_user from 8.8.8.8 port 12345 ssh2.",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "failed authentication",
                "authentication_type": "publickey",
                "user": "test_user",
                "remote_ip": "8.8.8.8",
                "port": 12345,
            },
            id="sshd: failed publickey",
        ),
        pytest.param(
            "Mar 27 13:06:56 ubuntu-1 sshd[1291]: Server listening on 127.0.0.1 port 22.",
            {
                "service": "sshd",
                "pid": 1291,
                "host_ip": "127.0.0.1",
                "port": 22,
            },
            id="sshd: listening",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: pam_unix(sshd:session): session opened for user test_user by (uid=0)",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "session opened",
                "user": "test_user",
                "user_uid": None,
                "by_uid": 0,
            },
            id="sshd: pam_unix",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: pam_unix(sshd:session): session opened "
            "for user root(uid=0) by (uid=0)",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "session opened",
                "user": "root",
                "user_uid": 0,
                "by_uid": 0,
            },
            id="sshd: pam_unix",
        ),
        pytest.param(
            "Mar 27 13:06:56 ubuntu-1 systemd-logind[1118]: Watching system buttons "
            "on /dev/input/event0 (Power Button)",
            {
                "service": "systemd-logind",
                "pid": 1118,
                "action": "Watching system buttons",
                "device": "/dev/input/event0",
                "device_name": "Power Button",
            },
            id="systemd-logind: watching system buttons",
        ),
        pytest.param(
            "Mar 27 13:06:56 ubuntu-1 systemd-logind[1118]: New seat seat0.",
            {
                "service": "systemd-logind",
                "pid": 1118,
                "action": "new seat",
                "seat": "seat0",
            },
            id="systemd-logind: new seat",
        ),
        pytest.param(
            "Mar 27 13:10:08 ubuntu-1 sudo:   ubuntu : TTY=pts/0 ; PWD=/home/test_user ; "
            "USER=root ; COMMAND=/usr/bin/apt-key add -",
            {
                "service": "sudo",
                "pid": None,
                "tty": "pts/0",
                "pwd": "/home/test_user",
                "effective_user": "root",
                "command": "/usr/bin/apt-key add -",
            },
            id="sudo: command",
        ),
        pytest.param(
            "Apr  3 12:32:23 ubuntu-1 su[1521]: Successful su for user by root",
            {"service": "su", "pid": 1521, "su_result": "success", "user": "user", "by": "root"},
            id="su: success",
        ),
        pytest.param(
            "Apr  3 12:32:23 ubuntu-1 su[1531]: 'su root' failed for user by root",
            {
                "service": "su",
                "pid": 1531,
                "su_result": "failed",
                "command": "su root",
                "user": "user",
                "by": "root",
            },
            id="su: failed",
        ),
        pytest.param(
            "Apr  3 12:32:23 ubuntu-1 pkexec[1531]: user: Executing command [USER=root] "
            "[TTY=unknown] [CWD=/home/user] [COMMAND=/usr/lib/update-notifier/package-system-locked]",
            {
                "service": "pkexec",
                "pid": 1531,
                "action": "executing command",
                "user": "user",
                "effective_user": "root",
                "tty": "unknown",
                "cwd": "/home/user",
                "command": "/usr/lib/update-notifier/package-system-locked",
            },
            id="pkexec: executing command",
        ),
        pytest.param(
            "Mar 27 13:17:01 ubuntu-1 CRON[2623]: pam_unix(cron:session): session closed for user root",
            {
                "service": "CRON",
                "pid": 2623,
                "action": "session closed",
                "user": "root",
            },
            id="cron: pam_unix",
        ),
    ],
)
def test_auth_additional_fields(
    target_unix: Target, fs_unix: VirtualFilesystem, tmp_path: Path, message: str, results: dict[str, str | int]
) -> None:
    data_path = tmp_path / "auth.log"
    data_path.write_text(message)
    fs_unix.map_file("var/log/auth.log", data_path)

    target_unix.add_plugin(AuthPlugin)

    result = list(target_unix.authlog())
    assert len(result) == 1

    for key, value in results.items():
        plugin_result = getattr(result[0], key)
        if isinstance(value, list):
            value = sorted(map(str, value))
            plugin_result = sorted(map(str, plugin_result))

        assert plugin_result == value


def test_auth_iso_date_format(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly handle Ubuntu 24.04 ISO formatted dates."""

    fs_unix.map_file("/var/log/auth.log", absolute_path("_data/plugins/os/unix/log/auth/iso.log"))
    target_unix.add_plugin(AuthPlugin)

    results = sorted(target_unix.authlog(), key=lambda r: r.ts)
    assert len(results) == 10

    assert results[0].ts == datetime(2024, 12, 31, 11, 37, 1, 123456, tzinfo=timezone.utc)
    assert results[0].service == "sudo"
    assert results[0].pid is None
    assert results[0].tty == "pts/0"
    assert results[0].pwd == "/home/user"
    assert results[0].effective_user == "root"
    assert results[0].command == "/usr/bin/chmod go+r /etc/apt/keyrings/githubcli-archive-keyring.gpg"
    assert results[0].source == "/var/log/auth.log"
    assert (
        results[0].message
        == "user : TTY=pts/0 ; PWD=/home/user ; USER=root ; COMMAND=/usr/bin/chmod go+r /etc/apt/keyrings/githubcli-archive-keyring.gpg"  # noqa: E501
    )


def test_auth_direct_mode() -> None:
    data_path = absolute_path("_data/plugins/os/unix/log/auth/auth.log")

    target = Target.open_direct([data_path])
    results = list(target.authlog())

    assert len(results) == 10



from __future__ import annotations

import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

import pytest
from dissect.util.ts import from_unix
from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.os.unix.history import CommandHistoryPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_commandhistory_with_timestamps(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    commandhistory_data = """\
    #1648598339
    echo "this is a test"
    echo "O no. A line without timestamp"
    #1658578468
    exit
    """

    commandhistory2_data = """\
    echo "this for user 2"
    """

    fs_unix.map_file_fh(
        "/root/.bash_history",
        BytesIO(textwrap.dedent(commandhistory_data).encode()),
    )

    fs_unix.map_file_fh(
        "/home/user/.bash_history",
        BytesIO(textwrap.dedent(commandhistory2_data).encode()),
    )

    target_unix_users.add_plugin(CommandHistoryPlugin)

    results = list(target_unix_users.commandhistory())
    assert len(results) == 4

    assert results[0].ts == dt("2022-03-29T23:58:59Z")
    assert results[0].command == 'echo "this is a test"'
    assert results[0].order == 0
    assert results[0].shell == "bash"
    assert results[0].source.as_posix() == "/root/.bash_history"

    assert results[1].ts is None
    assert results[1].command == 'echo "O no. A line without timestamp"'
    assert results[1].order == 1
    assert results[1].shell == "bash"
    assert results[1].source.as_posix() == "/root/.bash_history"

    assert results[2].ts == dt("2022-07-23T12:14:28Z")
    assert results[2].command == "exit"
    assert results[2].order == 2
    assert results[2].shell == "bash"
    assert results[2].source.as_posix() == "/root/.bash_history"

    assert results[3].ts is None
    assert results[3].command == 'echo "this for user 2"'
    assert results[3].order == 0
    assert results[3].shell == "bash"
    assert results[3].source.as_posix() == "/home/user/.bash_history"


def test_commandhistory_without_timestamps(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    commandhistory_data = """\
    echo "Test if basic commandhistory works" > /dev/null
    exit
    """

    fs_unix.map_file_fh(
        "/root/.zsh_history",
        BytesIO(textwrap.dedent(commandhistory_data).encode()),
    )

    target_unix_users.add_plugin(CommandHistoryPlugin)

    results = list(target_unix_users.commandhistory())
    assert len(results) == 2

    assert results[0].ts is None
    assert results[0].command == 'echo "Test if basic commandhistory works" > /dev/null'
    assert results[0].order == 0
    assert results[0].shell == "zsh"
    assert results[0].source.as_posix() == "/root/.zsh_history"

    assert results[1].ts is None
    assert results[1].command == "exit"
    assert results[1].order == 1
    assert results[1].shell == "zsh"
    assert results[1].source.as_posix() == "/root/.zsh_history"


def test_commandhistory_zsh_history(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    commandhistory_data = """\
    : 1673860722:0;sudo apt install sl
    : :;
    echo "Whoops no timestamps"
    """

    fs_unix.map_file_fh(
        "/root/.zsh_history",
        BytesIO(textwrap.dedent(commandhistory_data).encode()),
    )

    target_unix_users.add_plugin(CommandHistoryPlugin)

    results = list(target_unix_users.commandhistory())
    assert len(results) == 2

    assert results[0].ts == from_unix(1673860722)
    assert results[0].command == "sudo apt install sl"
    assert results[0].order == 0
    assert results[0].shell == "zsh"
    assert results[0].source.as_posix() == "/root/.zsh_history"

    assert not results[1].ts
    assert results[1].command == 'echo "Whoops no timestamps"'
    assert results[1].order == 1
    assert results[1].shell == "zsh"
    assert results[1].source.as_posix() == "/root/.zsh_history"


def test_commandhistory_fish_history(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    commandhistory_data = """
    - cmd: ls
      when: 1688642435
    - cmd: cd home/
      when: 1688642441
      paths:
        - home/
    - cmd: echo "test: test"
      when: 1688986629
    """

    fs_unix.map_file_fh(
        "/root/.local/share/fish/fish_history",
        BytesIO(textwrap.dedent(commandhistory_data).encode()),
    )

    target_unix_users.add_plugin(CommandHistoryPlugin)

    results = list(target_unix_users.commandhistory())
    assert len(results) == 3

    assert results[0].ts == from_unix(1688642435)
    assert results[0].command == "ls"
    assert results[0].order == 0
    assert results[0].shell == "fish"
    assert results[0].source.as_posix() == "/root/.local/share/fish/fish_history"

    assert results[1].ts == from_unix(1688642441)
    assert results[1].command == "cd home/"
    assert results[1].order == 1
    assert results[1].shell == "fish"
    assert results[1].source.as_posix() == "/root/.local/share/fish/fish_history"

    assert results[2].ts == from_unix(1688986629)
    assert results[2].command == 'echo "test: test"'
    assert results[2].order == 2
    assert results[2].shell == "fish"
    assert results[2].source.as_posix() == "/root/.local/share/fish/fish_history"


@pytest.mark.parametrize(
    ("db_type", "db_file", "db_history"),
    [
        (
            "mongodb",
            ".dbshell",
            """
            db
            print("Hello World")
            db.users.find({ name: /foo/i }, { _id: 0, name: 1})
            """,
        ),
        (
            "postgresql",
            ".psql_history",
            """
            \\l
            \\c example
            \\dt
            select * from users;
            """,
        ),
        (
            "mysql",
            ".mysql_history",
            """
            show databases;
            use example;
            show tables;
            select * from users;
            """,
        ),
        (
            "sqlite",
            ".sqlite_history",
            """
            .help
            create table tbl1(one text, two int);
            insert into tbl1 values('hello!', 10);
            insert into tbl1 values('goodbye', 20);
            select * from tbl1;
            """,
        ),
    ],
)
def test_commandhistory_database_history(
    target_unix_users: Target, fs_unix: VirtualFilesystem, db_type: str, db_file: str, db_history: str
) -> None:
    history_content = textwrap.dedent(db_history)
    history_lines = history_content.strip().split("\n")
    fs_unix.map_file_fh(
        f"/root/{db_file}",
        BytesIO(history_content.encode()),
    )

    target_unix_users.add_plugin(CommandHistoryPlugin)

    results = list(target_unix_users.commandhistory())

    assert len(results) == len(history_lines)

    for i, line in enumerate(history_lines):
        assert results[i].ts is None
        assert results[i].command == line
        assert results[i].order == i
        assert results[i].shell == db_type
        assert results[i].source.as_posix() == f"/root/{db_file}"


def test_commandhistory_is_directory(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    commandhistory_data = """test"""

    fs_unix.map_file_fh(
        "/root/.zsh_history",
        BytesIO(textwrap.dedent(commandhistory_data).encode()),
    )

    fs_unix.makedirs("/root/.bash_history")
    results = list(target_unix_users.commandhistory())

    assert len(results) == 1

    assert results[0].ts is None
    assert results[0].command == "test"
    assert results[0].order == 0
    assert results[0].shell == "zsh"
    assert results[0].source.as_posix() == "/root/.zsh_history"



from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING

from dissect.cim import cim
from dissect.util.sid import read_sid

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

# https://learn.microsoft.com/en-us/windows/win32/wmisdk/--eventconsumer
COMMON_ELEMENTS = [
    ("string", "creator_sid"),
    ("string", "machine_name"),
    ("string", "filter_query"),
    ("string", "filter_name"),
    ("string", "filter_query_language"),
    ("string", "filter_creator_sid"),
]

CommandLineEventConsumerRecord = TargetRecordDescriptor(
    "filesystem/windows/cim/consumerbinding/commandlineeventconsumer",
    [
        *COMMON_ELEMENTS,
        ("string", "command_line_template"),
        ("string", "executable_path"),
        ("string", "working_directory"),
    ],
)

# https://learn.microsoft.com/en-us/windows/win32/wmisdk/activescripteventconsumer
ActiveScriptEventConsumerRecord = TargetRecordDescriptor(
    "filesystem/windows/cim/consumerbinding/activescripteventconsumer",
    [
        *COMMON_ELEMENTS,
        ("string", "script_text"),
        ("string", "script_file_name"),
        ("string", "scripting_engine"),
        ("string", "name"),
    ],
)


@dataclass
class EventFilter:
    """Extracted information from ``__EventFilter``.

    References:
        - https://learn.microsoft.com/en-us/windows/win32/wmisdk/--eventfilter
    """

    filter_name: str = ""
    filter_query: str = ""
    filter_query_language: str = ""
    filter_creator_sid: str = ""


def get_property_value_safe(consumer: cim.Instance, prop_name: str, default_value: str | None = None) -> str | None:
    """Extract value of a consumer properties. Fallback to ``default_value`` if property is missing."""
    if not (prop := consumer.properties.get(prop_name)):
        return default_value
    try:
        return str(prop.value)
    except ValueError:
        return default_value


def get_filter_name(binding: cim.Instance) -> str:
    """Return unquoted filter name from a ``__filtertoconsumerbinding`` class instance."""
    filter_name = binding.properties["Filter"].value
    # filter name is not always consistent
    # e.g : __EventFilter.Name="Windows Update Event MOF" or
    # \\.\root\subscription:__EventFilter.Name="Windows Update Event MOF"
    if "=" in filter_name:
        # Required to manage filters name with escaped "
        _, _, filter_name = filter_name.partition("=")
        filter_name = filter_name.strip('"').replace('\\"', '"')
    return filter_name


def get_creator_sid(class_instance: cim.Instance) -> str | None:
    """Extract and parse ``CreatorSID`` member, if available."""
    if (creator_sid := class_instance.properties.get("CreatorSID")) and (
        creator_sid_value := getattr(creator_sid, "value", None)
    ):
        return read_sid(bytes(creator_sid_value))
    return None


class CimPlugin(Plugin):
    """CIM database plugin.

    Provides functions for getting useful data out the CIM (WBEM) database.
    """

    __namespace__ = "cim"

    def __init__(self, target: Target):
        super().__init__(target)
        self._repo = None
        repodir = self.target.resolve("%windir%/system32/wbem/repository")
        self._subscription_ns = None
        self._filters: dict[str, EventFilter] = {}
        if repodir.exists():
            index = repodir.joinpath("index.btr")
            objects = repodir.joinpath("objects.data")
            mappings = [repodir.joinpath(f"mapping{i}.map") for i in range(1, 4)]

            if all([index.exists(), objects.exists(), all(m.exists() for m in mappings)]):
                try:
                    self._repo = cim.CIM(index.open(), objects.open(), [m.open() for m in mappings])
                except cim.Error as e:
                    self.target.log.warning("Error opening CIM database")
                    self.target.log.debug("", exc_info=e)
            self._subscription_ns = self._repo.root.namespace("subscription")
            self._filters = self._get_filters()

    def check_compatible(self) -> None:
        if not self._repo:
            raise UnsupportedPluginError("No WBEM repository found")

    @internal
    def repo(self) -> cim.CIM:
        return self._repo

    def _iter_consumerbindings(self) -> Iterator[tuple[cim.Instance, str]]:
        """Yield consumer bindings from ``__filtertoconsumerbinding`` of subscription namespace."""
        try:
            for binding in self._subscription_ns.class_("__filtertoconsumerbinding").instances:
                yield (
                    self._subscription_ns.query(binding.properties["Consumer"].value),
                    get_filter_name(binding),
                )
        except Exception as e:
            self.target.log.warning("Error retrieving consumerbindings")
            self.target.log.debug("", exc_info=e)

    @export(record=[ActiveScriptEventConsumerRecord, CommandLineEventConsumerRecord])
    def consumerbindings(self) -> Iterator[ActiveScriptEventConsumerRecord | CommandLineEventConsumerRecord]:
        """Return all ActiveScriptEventConsumer and CommandLineEventConsumer.

        WMI permanent event subscriptions can be used to trigger actions when specified conditions are met. Attackers
        often use this functionality to persist the execution of backdoors at system start up. WMI Consumers specify an
        action to be performed, including executing a command, running a script, adding an entry to a log, or sending
        an email. WMI Filters define conditions that will trigger a Consumer.

        References:
            - https://learn-powershell.net/2013/08/14/powershell-and-events-permanent-wmi-event-subscriptions/
            - https://www.mandiant.com/resources/dissecting-one-ofap
            - https://support.sophos.com/support/s/article/KB-000038535?language=en_US&c__displayLanguage=en_US
            - https://learn.microsoft.com/en-us/windows/win32/wmisdk/activescripteventconsumer
            - https://learn.microsoft.com/en-us/windows/win32/wmisdk/commandlineeventconsumer
        """
        for consumer, filter_name in self._iter_consumerbindings():
            if consumer.properties.get("ScriptText"):
                yield ActiveScriptEventConsumerRecord(
                    script_text=get_property_value_safe(consumer, "ScriptText", ""),
                    script_file_name=get_property_value_safe(consumer, "ScriptFileName", ""),
                    scripting_engine=get_property_value_safe(consumer, "ScriptingEngine", ""),
                    machine_name=get_property_value_safe(consumer, "MachineName", ""),
                    name=get_property_value_safe(consumer, "Name", ""),
                    creator_sid=get_creator_sid(consumer),
                    _target=self.target,
                    **asdict(self._filters.get(filter_name, EventFilter(filter_name=filter_name))),
                )
            if consumer.properties.get("CommandLineTemplate"):
                yield CommandLineEventConsumerRecord(
                    command_line_template=get_property_value_safe(consumer, "CommandLineTemplate", ""),
                    executable_path=get_property_value_safe(consumer, "ExecutablePath", ""),
                    working_directory=get_property_value_safe(consumer, "WorkingDirectory", ""),
                    creator_sid=get_creator_sid(consumer),
                    _target=self.target,
                    **asdict(self._filters.get(filter_name, EventFilter(filter_name=filter_name))),
                )

    def _get_filters(self) -> dict[str, EventFilter]:
        """Generate a dictionary of ``__EventFilter`` that can be mapped with ``__filtertoconsumerbinding``."""
        filters = {}
        for event in self._subscription_ns.class_("__EventFilter").instances:
            filter_name = event.properties["Name"].value
            filters[filter_name] = EventFilter(
                filter_name=filter_name,
                filter_query=event.properties["Query"].value,
                filter_query_language=event.properties["QueryLanguage"].value,
                filter_creator_sid=get_creator_sid(event),
            )
        return filters



from __future__ import annotations

import argparse
import logging

from dissect.util.feature import Feature, feature_enabled

from dissect.target import filesystem
from dissect.target.exceptions import TargetError
from dissect.target.helpers.utils import parse_options_string
from dissect.target.target import Target
from dissect.target.tools.utils import (
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)

# Setting logging level to info for startup information.
logging.basicConfig(level=logging.INFO)

try:
    if feature_enabled(Feature.BETA):
        from fuse3 import FUSE3 as FUSE
        from fuse3 import util

        FUSE_VERSION = "3"
        FUSE_LIB_PATH = util.libfuse._name
    else:
        from fuse import FUSE, _libfuse

        FUSE_VERSION = "2"
        FUSE_LIB_PATH = _libfuse._name

    logging.info("Using fuse%s library: %s", FUSE_VERSION, FUSE_LIB_PATH)

    from dissect.target.helpers.mount import DissectMount

    HAS_FUSE = True
except Exception:
    HAS_FUSE = False


log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("target", metavar="TARGET", help="target to load")
    parser.add_argument("mount", metavar="MOUNT", help="path to mount to")
    parser.add_argument("-o", "--options", help="additional FUSE options")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(args)

    if not HAS_FUSE:
        log.error("fusepy is not installed: pip install fusepy")
        return 1

    try:
        t = Target.open(args.target)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    vfs = filesystem.VirtualFilesystem()
    vfs.mount("fs", t.fs)

    for i, d in enumerate(t.disks):
        fname = f"disks/disk_{i}"
        vfs.map_file_fh(fname, d)

    vnames = {}
    for i, v in enumerate(t.volumes):
        basename = f"{v.name or f'volume_{i}'}"
        fname = basename

        j = 1
        while fname in vnames:
            fname = f"{basename}_{j}"
            j += 1

        vnames[v] = fname
        vnames.setdefault(basename, []).append(v)
        vfs.map_file_fh(f"volumes/{fname}", v)

    for i, fs in enumerate(t.filesystems):
        volumes = fs.volume if isinstance(fs.volume, list) else [fs.volume]
        for volume in volumes:
            default_name = f"fs_{i}"
            fname = f"filesystems/{vnames.get(volume, default_name) if volume else default_name}"
            vfs.mount(fname, fs)

    # This is kinda silly because fusepy will convert this back into string arguments
    options = parse_options_string(args.options) if args.options else {}

    options["nothreads"] = True
    options["ro"] = True
    # Check if the allow other option is either not set (None) or set to True with -o allow_other=True
    if (allow_other := options.get("allow_other")) is None or str(allow_other).lower() == "true":
        options["allow_other"] = True
        # If allow_other was not set, warn the user that it will be set by default
        if allow_other is None:
            log.warning("Using option 'allow_other' by default, please use '-o allow_other=False' to unset")
    # Let the user be able to unset the allow_other option by supplying -o allow_other=False
    elif str(allow_other).lower() == "false":
        options["allow_other"] = False

    log.info("Mounting to %s with options: %s", args.mount, _format_options(options))
    try:
        FUSE(DissectMount(vfs), args.mount, **options)
    except RuntimeError as e:
        log.error("Mounting target %s failed", t)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    return 0


def _format_options(options: dict[str, str | bool]) -> str:
    return ",".join(key if value is True else f"{key}={value}" for key, value in options.items())


if __name__ == "__main__":
    main()



from __future__ import annotations

import struct
from typing import TYPE_CHECKING

from dissect.util.ts import wintimestamp

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import TargetRecordDescriptor, create_extended_descriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.regf.shellbags import (
    FILE_ENTRY,
    parse_shell_item_list,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record import Record

    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.target import Target

UserRegistryRecordDescriptor = create_extended_descriptor(
    [
        RegistryRecordDescriptorExtension,
        UserRecordDescriptorExtension,
    ]
)

RunMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/run",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

RecentDocsRecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/recentdocs",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

OpenSaveMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/opensave",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

LastVisitedMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/lastvisited",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "filename"),
        ("string", "path"),
        ("string", "key"),
    ],
)

ACMruRecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/acmru",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "category"),
        ("string", "value"),
        ("string", "key"),
    ],
)

MapNetworkDriveMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/mapnetworkdrive",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

TerminalServerMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/mstsc",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

MSOfficeMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/msoffice",
    [
        ("datetime", "ts"),
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)


class MRUPlugin(Plugin):
    """Return MRU data stored at various registry keys.

    The Windows registry contains various keys about Most Recently Used (MRU) files.

    References:
        - https://winreg-kb.readthedocs.io/en/latest/sources/explorer-keys/Most-recently-used.html
    """

    __namespace__ = "mru"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("Target has no registry")

    @export(record=RunMRURecord)
    def run(self) -> Iterator[RunMRURecord]:
        """Return the RunMRU data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RunMRU`` registry key contains information
        about the most recent commands that have been performed by the Run application

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """
        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RunMRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, RunMRURecord)

    @export(record=RecentDocsRecord)
    def recentdocs(self) -> Iterator[RecentDocsRecord]:
        """Return the RecentDocs data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RecentDocs`` registry key contains
        information about the last 10 documents that the currently logged on user accessed or executed via Windows
        Explorer.

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RecentDocs"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_ex_key(self.target, key, RecentDocsRecord)

    @export(record=OpenSaveMRURecord)
    def opensave(self) -> Iterator[OpenSaveMRURecord]:
        """Return the OpenSaveMRU data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSaveMRU`` registry key
        - renamed to ``OpenSavePidlMRU`` since Windows Vista - contains information about the most recently
        opened or saved files.

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSaveMRU"
        PIDL_KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSavePidlMRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, OpenSaveMRURecord)

        for key in self.target.registry.keys(PIDL_KEY):
            yield from parse_mru_ex_key(self.target, key, OpenSaveMRURecord)

    @export(record=LastVisitedMRURecord)
    def lastvisited(self) -> Iterator[LastVisitedMRURecord]:
        """Return the LastVisitedMRU data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedMRU`` registry key
        - renamed to ``LastVisitedPidlMRU`` since Windows Vista - contains information about the executable used by
        an application to open the files that are documented at the ``OpenSaveMRU`` registry key. Also, each value
        tracks the directory location for the last file that was accessed by that application.

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedMRU"
        PIDL_KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedPidlMRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, LastVisitedMRURecord)

        for key in self.target.registry.keys(PIDL_KEY):
            yield from parse_mru_ex_key(self.target, key, LastVisitedMRURecord)

    @export(record=ACMruRecord)
    def acmru(self) -> Iterator[ACMruRecord]:
        """Return the ACMru (Windows Search) data.

        The following keys are being searched:
          - ``HKCU\\Software\\Microsoft\\Search Assistant\\ACMru``:
            This registry key contains the most recent search history from Windows default search.
          - ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\WordWheelQuery``:
            This registry key contains the most recent search history from Windows Explorer. (Windows >=7)

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/

        Known categories:
            - 5001: Internet Search Assistant
            - 5603: Windows XP files and folder search
            - 5604: "Word or phrase in a file" dialog box
            - 5647: "For computers or people" selection in Search Results dialog box
        """

        KEY = "HKCU\\Software\\Microsoft\\Search Assistant\\ACMru"

        for key in self.target.registry.keys(KEY):
            user = self.target.registry.get_user(key)
            for subkey in key.subkeys():
                for value in subkey.values():
                    yield ACMruRecord(
                        regf_mtime=key.ts,
                        index=int(value.name),
                        category=subkey.name,
                        value=value.value,
                        _target=self.target,
                        _user=user,
                        _key=subkey,
                    )

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\WordWheelQuery"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_ex_key(self.target, key, ACMruRecord)

    @export(record=MapNetworkDriveMRURecord)
    def networkdrive(self) -> Iterator[MapNetworkDriveMRURecord]:
        """Return MRU of mapped network drives.

        The HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Map Network Drive MRU registry key contains
        information about the most recently used mapped network drives.

        References:
            - https://winreg-kb.readthedocs.io/en/latest/sources/explorer-keys/Most-recently-used.html#keys-with-a-mrulist-value
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Map Network Drive MRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, MapNetworkDriveMRURecord)

    @export(record=TerminalServerMRURecord)
    def mstsc(self) -> Iterator[TerminalServerMRURecord]:
        """Return Terminal Server Client MRU data."""

        KEY = "HKCU\\Software\\Microsoft\\Terminal Server Client\\Default"

        for key in self.target.registry.keys(KEY):
            user = self.target.registry.get_user(key)

            for value in key.values():
                entry_index = int(value.name.split("MRU")[1])

                yield TerminalServerMRURecord(
                    regf_mtime=key.ts,
                    index=entry_index,
                    value=value.value,
                    _target=self.target,
                    _user=user,
                    _key=key,
                )

    @export(record=MSOfficeMRURecord)
    def msoffice(self) -> Iterator[MSOfficeMRURecord]:
        """Return MS Office MRU keys."""

        KEY = "HKCU\\Software\\Microsoft\\Office"
        SUBKEYS = [
            "Common",
            "Excel",
            "Groove",
            "OneNote",
            "Outlook",
            "PowerPoint",
            "Publisher",
            "Word",
        ]

        for key in self.target.registry.keys(KEY):
            for version_key in key.subkeys():
                if not version_key.name[0].isdigit():
                    continue

                for subkey in version_key.subkeys():
                    if subkey.name not in SUBKEYS:
                        continue

                    try:
                        yield from parse_office_mru(self.target, subkey, MSOfficeMRURecord)
                    except RegistryError:
                        pass


def parse_mru_key(target: Target, key: RegistryKey, record: TargetRecordDescriptor) -> Iterator[Record]:
    user = target.registry.get_user(key)

    try:
        mrulist = key.value("MRUList").value
    except RegistryError:
        mrulist = None

    for value in key.values():
        if value.name == "MRUList":
            continue

        if record == LastVisitedMRURecord:
            entry_index = mrulist.index(value.name) if mrulist else None
            values = read_wstrings(value.value)

            yield record(
                regf_mtime=key.ts,
                index=entry_index,
                filename=next(values).decode("utf-16-le"),
                path=next(values).decode("utf-16-le"),
                _target=target,
                _user=user,
                _key=key,
            )
        else:
            entry_index = mrulist.index(value.name) if mrulist else None
            entry_value = value.value

            yield record(
                regf_mtime=key.ts,
                index=entry_index,
                value=entry_value,
                _target=target,
                _user=user,
                _key=key,
            )

    for subkey in key.subkeys():
        yield from parse_mru_key(target, subkey, record)


def parse_mru_ex_key(target: Target, key: RegistryKey, record: TargetRecordDescriptor) -> Iterator[Record]:
    user = target.registry.get_user(key)

    try:
        mrulist_ex = key.value("MRUListEx").value
        mrulist_ex = struct.unpack(f"<{len(mrulist_ex) // 4}I", mrulist_ex)
    except RegistryError:
        mrulist_ex = None

    for value in key.values():
        if value.name == "MRUListEx":
            continue

        entry_index = mrulist_ex.index(int(value.name))
        if record != OpenSaveMRURecord:
            path = next(read_wstrings(value.value))
            bag = value.value[len(path) + 2 :]
        else:
            bag = value.value
            path = None
        parsed_bag = list(parse_shell_item_list(bag))
        if len(parsed_bag) != 1 or not isinstance(parsed_bag, FILE_ENTRY):
            target.log.debug("Unexpected shell bag entry in MRUListEx entry: %s:%s", key, value)

        if record == LastVisitedMRURecord:
            filepath = "\\".join(el.name.rstrip("\\") for el in parsed_bag) + "\\"
            yield record(
                regf_mtime=key.ts,
                index=entry_index,
                filename=path.decode("utf-16-le"),
                path=filepath,
                _target=target,
                _user=user,
                _key=key,
            )
        elif record == OpenSaveMRURecord:
            filepath = "\\".join(el.name.rstrip("\\") for el in parsed_bag)
            yield record(
                regf_mtime=key.ts,
                index=entry_index,
                value=filepath,
                _target=target,
                _user=user,
                _key=key,
            )
        else:
            yield record(
                regf_mtime=key.ts,
                index=entry_index,
                value=path.decode("utf-16-le"),
                _target=target,
                _user=user,
                _key=key,
            )

    for subkey in key.subkeys():
        yield from parse_mru_ex_key(target, subkey, record)


def parse_office_mru(target: Target, key: RegistryKey, record: TargetRecordDescriptor) -> Iterator[Record]:
    try:
        yield from parse_office_mru_key(target, key.subkey("File MRU"), record)
    except RegistryError:
        pass

    try:
        yield from parse_office_mru_key(target, key.subkey("Place MRU"), record)
    except RegistryError:
        pass

    try:
        for subkey in key.subkey("User MRU").subkeys():
            yield from parse_office_mru(target, subkey, record)
    except RegistryError:
        pass


def parse_office_mru_key(target: Target, key: RegistryKey, record: TargetRecordDescriptor) -> Iterator[Record]:
    user = target.registry.get_user(key)

    for value in key.values():
        if not value.name.startswith("Item"):
            continue

        entry_index = int(value.name.split(" ")[1])
        info_str, path = value.value.split("*", 1)

        info = {part[0]: int(part[1:], 16) for part in info_str.strip("[]").split("][")}

        yield record(
            ts=wintimestamp(info["T"]),
            regf_mtime=key.ts,
            index=entry_index,
            value=path,
            _target=target,
            _user=user,
            _key=key,
        )


def read_wstrings(buf: bytes) -> Iterator[bytes]:
    idx = 0
    remaining = buf
    while (idx := remaining.find(b"\x00\x00", idx)) != -1:
        if idx & 1:
            idx += 1
            continue

        yield remaining[:idx]
        remaining = remaining[idx + 2 :]
        idx = 0

    if remaining:
        yield remaining



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.child.hyperv import HyperVChildTargetPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_wsl(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "ProgramData/Microsoft/Windows/Hyper-V/data.vmcx",
        absolute_path("_data/plugins/child/hyperv/data.vmcx"),
    )
    fs_win.map_file(
        "ProgramData/Microsoft/Windows/Hyper-V/Virtual Machines/B90AC31B-C6F8-479F-9B91-07B894A6A3F6.xml",
        absolute_path("_data/loaders/hyperv/B90AC31B-C6F8-479F-9B91-07B894A6A3F6.xml"),
    )

    target_win.add_plugin(HyperVChildTargetPlugin)

    children = list(target_win.list_children())

    assert len(children) == 5
    assert (
        str(children[0].path) == "C:\\Hyper-V\\EasyToFind\\Virtual Machines\\EC04F346-DB96-4700-AF5B-77B3C56C38BD.vmcx"
    )
    assert (
        str(children[1].path) == "C:\\Hyper-V\\EasyToFind\\Virtual Machines\\993F7B33-6057-4D1E-A1FE-A1A1D77BE974.vmcx"
    )
    assert (
        str(children[2].path)
        == "C:\\VM\\Other Generation 1\\Virtual Machines\\A5B56431-CA94-482A-B70A-F1F2B12373BE.vmcx"
    )
    assert (
        str(children[3].path)
        == "C:\\VM\\Other Generation 2\\Virtual Machines\\4C57771A-3230-4B92-B029-D63F96518E70.vmcx"
    )
    assert (
        str(children[4].path)
        == "\\sysvol\\ProgramData\\Microsoft\\Windows\\Hyper-V\\Virtual Machines\\B90AC31B-C6F8-479F-9B91-07B894A6A3F6.xml"  # noqa E501
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.helpers.localeutil import normalize_language
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.default.locale import LocalePlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

UnixKeyboardRecord = TargetRecordDescriptor(
    "unix/keyboard",
    [
        ("string", "layout"),
        ("string", "model"),
        ("string", "variant"),
        ("string", "options"),
        ("string", "backspace"),
    ],
)


def timezone_from_path(path: Path | str) -> str:
    """Return timezone name for the given zoneinfo path.

    .. code-block::

        /usr/share/zoneinfo/Europe/Amsterdam -> Europe/Amsterdam
        /usr/share/zoneinfo/UTC              -> UTC
        Etc/UTC                              -> UTC
    """
    if not isinstance(path, Path):
        path = Path(path)

    return "/".join([p for p in path.parts[-2:] if p.lower() not in ["zoneinfo", "etc"]])


class UnixLocalePlugin(LocalePlugin):
    """Unix locale plugin."""

    def check_compatible(self) -> None:
        pass

    @export(property=True)
    def timezone(self) -> str | None:
        """Get the timezone of the system."""

        # /etc/timezone should contain a simple timezone string
        # on most unix systems
        if (path := self.target.fs.path("/etc/timezone")).exists():
            for line in path.open("rt"):
                return timezone_from_path(Path(line.strip()))

        # /etc/localtime can be a symlink, hardlink or a copy of:
        # eg. /usr/share/zoneinfo/America/New_York
        p_localtime = self.target.fs.path("/etc/localtime")

        # If it's a symlink, read the path of the symlink
        if p_localtime.is_symlink():
            return timezone_from_path(p_localtime.readlink())

        # If it's a hardlink, try finding the hardlinked zoneinfo file
        if p_localtime.exists() and p_localtime.stat().st_nlink > 1:
            for path in self.target.fs.path("/usr/share/zoneinfo").rglob("*"):
                if p_localtime.samefile(path):
                    return timezone_from_path(path)

        # If it's a regular file (probably copied), we try finding it by matching size and sha1 hash.
        if p_localtime.is_file():
            size = p_localtime.stat().st_size
            sha1 = p_localtime.get().sha1()
            for path in self.target.fs.path("/usr/share/zoneinfo").rglob("*"):
                # Ignore posix files in zoneinfo directory (RHEL).
                if path.name.startswith("posix"):
                    continue

                if path.is_file() and path.stat().st_size == size and path.get().sha1() == sha1:
                    return timezone_from_path(path)
            return None
        return None

    @export(property=True)
    def language(self) -> list[str]:
        """Get the configured locale(s) of the system."""

        # Although this purports to be a generic function for Unix targets, these paths are Linux specific.
        locale_paths = [
            "/etc/default/locale",
            "/etc/locale.conf",
            "/etc/sysconfig/i18n",
        ]

        found_languages = set()

        for locale_path in locale_paths:
            if (path := self.target.fs.path(locale_path)).exists():
                for line in path.open("rt"):
                    if "LANG=" in line:
                        lang_str = line.partition("=")[-1].strip().strip('"')
                        if lang_str == "C.UTF-8":  # Skip if no locales are installed.
                            continue
                        found_languages.add(normalize_language(lang_str))

        return list(found_languages)

    @export(record=UnixKeyboardRecord)
    def keyboard(self) -> Iterator[UnixKeyboardRecord]:
        """Get the keyboard layout(s) of the system."""

        paths = [
            "/etc/default/keyboard",
            "/etc/vconsole.conf",
            *list(self.target.fs.glob("/etc/X11/xorg.conf.d/*-keyboard.conf")),
        ]

        for path_ in paths:
            if (path := self.target.fs.path(path_)).exists():
                k = {}
                for line in path.open("rt"):
                    if len(key_value := line.split("=")) == 2:
                        k[key_value[0].replace("XKB", "").replace("KEYMAP", "LAYOUT")] = key_value[1].strip().strip('"')

                yield UnixKeyboardRecord(
                    layout=k.get("LAYOUT"),
                    model=k.get("MODEL"),
                    variant=k.get("VARIANT"),
                    options=k.get("OPTIONS"),
                    backspace=k.get("BACKSPACE"),
                    _target=self.target,
                )

        # TODO
        # /etc/sysconfig/keyboard
        # /usr/share/kbd/keymaps/*
        # /usr/local/share/kbd/keymaps/*



from __future__ import annotations

import re
from random import randrange
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest
from dissect.ntfs.c_ntfs import ATTRIBUTE_TYPE_CODE
from dissect.ntfs.exceptions import Error

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.plugins.filesystem.ntfs.mft import (
    FilesystemFilenameRecord,
    FilesystemStdRecord,
    MftPlugin,
    macb_aggregator,
)
from dissect.target.plugins.filesystem.ntfs.utils import (
    get_drive_letter,
    get_owner_and_group,
    get_volume_identifier,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target


@pytest.fixture(params=[True, False])
def compact(request: pytest.FixtureRequest) -> bool:
    return request.param


@pytest.fixture
def target_win_mft(target_win: Target) -> Target:
    filesystem = NtfsFilesystem(mft=absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw").open("rb"))
    target_win.filesystems = [filesystem]
    target_win.add_plugin(MftPlugin)
    return target_win


@pytest.mark.parametrize(
    ("drive_letters", "expected"),
    [
        (("z:", "c:", "a:"), "a:\\"),
        (("foo", "bar", "bla"), "bar\\"),
        (("$fs$\\fs2", "$fs$\\fs1", "$fs$\\fs0"), "$fs$\\fs0\\"),
        (("sysvol", "c:"), "c:\\"),
        (("sysvol", "c:", "a:"), "a:\\"),
        (("$fs$\\fs0", "sysvol", "c:"), "c:\\"),
        (("$fs$\\fs0", "sysvol"), "sysvol\\"),
        (("$fs$\\fs0",), "$fs$\\fs0\\"),
    ],
)
def test_driveletter(drive_letters: Iterator[str], expected: str) -> None:
    mocked_target = Mock()
    mocked_disk = Mock()
    num_drives = len(drive_letters)
    mocked_target.fs.mounts = dict(zip(drive_letters, [mocked_disk] * num_drives))

    assert get_drive_letter(mocked_target, mocked_disk) == expected


def test_driveletter_unknown() -> None:
    mocked_target = Mock()
    mocked_cdisk = Mock()
    mocked_target.fs.mounts = {"sysvol": mocked_cdisk, "c:": mocked_cdisk}
    assert get_drive_letter(mocked_target, Mock()) == ""


def test_driveletter_virtual_filesystems() -> None:
    mocked_target = Mock()
    mocked_ntfs = Mock()
    mock_fs = VirtualFilesystem()
    mock_fs.ntfs = mocked_ntfs
    mocked_ntfs_fs = Mock(spec=NtfsFilesystem)
    mocked_ntfs_fs.ntfs = mocked_ntfs
    mocked_target.fs.mounts = {"sysvol": VirtualFilesystem(), "c:": mock_fs}
    assert get_drive_letter(mocked_target, mocked_ntfs_fs) == "c:\\"


@pytest.mark.parametrize(
    "exception",
    [
        Error,
        AttributeError,
        KeyError,
        TypeError,
    ],
)
def test_get_owner_and_group_fail(exception: Exception) -> None:
    mocked_fs = Mock()
    mocked_fs.ntfs.secure.lookup.side_effect = [exception]

    mocked_entry = Mock()
    mocked_entry.attributes = {ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION: [Mock()]}
    assert get_owner_and_group(mocked_entry, mocked_fs) == (None, None)


def test_get_owner_and_group_none_attr() -> None:
    mocked_entry = Mock()
    mocked_entry.attributes = {ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION: [None]}
    assert get_owner_and_group(mocked_entry, Mock()) == (None, None)


@pytest.mark.parametrize(
    ("guid_bytes", "expected_result"),
    [
        (b",\xa6\xc1}\x88\xc4\xc5G\x8e\xab\t$\x0f8\x89N", "7dc1a62c-c488-47c5-8eab-09240f38894e"),
        (b"", None),
        (None, None),
    ],
)
def test_volume_identifier(guid_bytes: bytes, expected_result: str | None) -> None:
    filesystem = Mock()
    filesystem.volume.guid = guid_bytes
    assert get_volume_identifier(filesystem) == expected_result


def test_volume_identifier_no_volume() -> None:
    filesystem = Mock()
    filesystem.volume = None
    assert get_volume_identifier(filesystem) is None


def test_volume_identifier_none_guid() -> None:
    filesystem = Mock()
    filesystem.volume.guid = None
    assert get_volume_identifier(filesystem) is None


@pytest.mark.parametrize(
    ("regex_pattern", "expected_nr_of_records"),
    [
        (r".*", 304),
        (r".+Size:(No Data)", 120),
        (r".+Is_ADS$", 24),
        (r".+Size:\d+", 184),
        (
            r"2021-08-04 ([+.:]?\d+)+ [SF]0?[BCMA] 43 NamelessDirectory - "
            r"In[uU]se:True Resident:No Data Owner:No Data Size:No Data VolumeUUID:No Data",
            8,
        ),
        (
            r"2021-08-04 ([+.:]?\d+)+ [SF]0?[BCMA] 46 NamelessDirectory\\totally_normal.txt - "
            r"In[uU]se:True Resident:True Owner:No Data Size:28 VolumeUUID:No Data",
            8,
        ),
        (
            r"2021-08-04 ([+.:]?\d+)+ [SF]0?[BCMA] 46 NamelessDirectory\\totally_normal.txt"
            r":secret_text.txt - InUse:True Resident:True Owner:No Data Size:24 VolumeUUID:No Data Is_ADS",
            4,
        ),
    ],
)
def test_mft_timeline(target_win_mft: Target, regex_pattern: str, expected_nr_of_records: int) -> None:
    """Test whether the MFT timeline functions as inteded, same with the output."""
    outputs = [entry for entry in target_win_mft.mft.timeline() if re.match(regex_pattern, entry)]

    # Check the MFT entries
    assert len(outputs) == expected_nr_of_records


def check_output_amount(number: int, compact_output: bool) -> int:
    more_records = (0 if compact_output else 1) * 3
    return number + number * more_records


def test_mft_plugin_entries(target_win_mft: Target, compact: bool) -> None:
    mft_data = list(target_win_mft.mft(compact))
    assert len(mft_data) == check_output_amount(76, compact)


def test_mft_plugin_macb(target_win_mft: Target) -> None:
    mft_data = list(target_win_mft.mft(macb=True))
    path = None
    ts = None
    macb = None
    field = "MACB/MACB"
    for record in mft_data:
        assert record.macb != macb or record.ts != ts or record.path != path
        for bit in [0, 1, 2, 3, 5, 6, 7]:
            assert record.macb[bit : bit + 1] in (field[bit : bit + 1], ".")
        path = record.path
        macb = record.macb
        ts = record.ts


def test_mft_plugin_macb_ads(target_win_mft: Target) -> None:
    mft_data = list(target_win_mft.mft(macb=True))
    ads_entries = 0
    for record in mft_data:
        if record.ads:
            ads_entries += 1
            assert record.macb.endswith("/....")
            assert not record.macb.startswith("..../")
    assert ads_entries == 6


def test_mft_plugin_macb_nodup() -> None:
    # test whether you can never have duplicates

    def make_ts(tss: set) -> int:
        ts = randrange(1, 10)
        tss.add(ts)
        return ts

    for _ in range(100):
        tss = set()
        records = []
        for ts_type in ["M", "A", "C", "B"]:
            records.append(FilesystemStdRecord(path="a.txt", ts=make_ts(tss), ts_type=ts_type))
            records.append(FilesystemFilenameRecord(path="a.txt", ts=make_ts(tss), ts_type=ts_type, ads=False))
            records.append(FilesystemFilenameRecord(path="a.txt", ts=make_ts(tss), ts_type=ts_type, ads=True))
        assert len(list(macb_aggregator(records))) == len(tss)


def test_mft_plugin_disk_label(target_win_mft: Target) -> None:
    target_win_mft.fs.mounts = {"c:": target_win_mft.filesystems[0]}
    for mft_entries in target_win_mft.mft():
        assert str(mft_entries.path).startswith("c:\\")


def test_mft_plugin_ads(target_win_mft: Target, compact: bool) -> None:
    mft_data = [mft_entry for mft_entry in target_win_mft.mft(compact) if hasattr(mft_entry, "ads") and mft_entry.ads]
    assert len(mft_data) == check_output_amount(6, compact)


def test_mft_plugin_resident(target_win_mft: Target, compact: bool) -> None:
    mft_data = [mft_entry for mft_entry in target_win_mft.mft(compact) if mft_entry.resident]
    assert len(mft_data) == check_output_amount(19, compact)


def test_mft_plugin_inuse(target_win_mft: Target, compact: bool) -> None:
    mft_data = [mft_entry for mft_entry in target_win_mft.mft(compact) if mft_entry.inuse]
    assert len(mft_data) == check_output_amount(76, compact)


def test_mft_plugin_last_entries(target_win_mft: Target) -> None:
    mft_data = list(target_win_mft.mft())[-9:]
    test_data = [
        "NamelessDirectory",
        "Food For Thought",
        "text_document.txt",
        "NamelessDirectory\\totally_normal.txt",
        "NamelessDirectory\\totally_normal.txt:secret_text.txt",
    ]
    for mft_entry in mft_data:
        assert str(mft_entry.path) in test_data


def test_mft_plugin_owner(target_win_mft: Target) -> None:
    for mft_entry in target_win_mft.mft():
        assert mft_entry.owner is None
